#ifndef SI_ATOM
#define SI_ATOM


#include "SIarithmetic.h"
#include "myvector.h"

// Definition of comparison operator type
enum SI_operator{gt,geq,lt,leq,eq,neq,unset};

class SIatom {

 private:

  // tells whether the variable is already in the atom variables list
  unsigned int varIn(SIvariable t);

 protected:

  // indicates whether the atom is identically true
  unsigned int istrue;

  // indicates whether the atom is identically false
  unsigned int isfalse;

  // arithmetic formulas composing the atom
  SIarithmetic left;
  SIarithmetic right;

  // comparison operator
  SI_operator op;

  // variables contained in the atom
  myvector<SIvariable> varlist;

 public:

  SIatom(void);

  SIatom(SIarithmetic leftexp, SIarithmetic rightexp, SI_operator opr);

  // A call to this method destroys the old atom an creates a new one
  void setAtom(SIarithmetic leftexp, SIarithmetic rightexp, SI_operator opr);

  // reallineates the variables list according to the variables
  // contained in the arithmetic expressions
  void allineate(void);

  // copy constructor
  SIatom(const SIatom& a);

  const SIatom & operator = (const SIatom &a);

  // creates a FALSE atom if logicval==0, and a TRUE atom for all 
  // other possible values of logicval
  void setLogicValue(unsigned int logicval);

  // returns all the variables contained in the atom in a list
  myvector<SIvariable> getVariables(void);

  // replaces all occurrences of the variable whose name is in oldvar
  // with the arithmetic expression newexpr
  // returns a positive integer if at least an occurrence of the 
  // variable whose name is in oldvar has been found in the formula, 
  // returns 0 otherwise (no substitution has been performed).
  unsigned int replace(char *oldvar,SIarithmetic newexpr);

  // replaces all occurrences of the variable whose name is in oldvar
  // with the variable name contained in newvar
  // if newvar is not a number, it is added to the current variables list with
  // the same type as the replaced variable(oldvar)
  // returns a positive integer if at least an occurrence of the 
  // variable whose name is in oldvar has been found in the formula, 
  // returns 0 otherwise (no substitution has been performed).
  unsigned int replace(char *oldvar,char *newvar);

  // replaces only the first occurrence of the variable whose name is in oldvar
  // with the variable name contained in newvar
  // if newvar is not a number, it is added to the current variables list with
  // the same type as the replaced variable(oldvar)
  // returns a positive integer if at least an occurrence of the 
  // variable whose name is in oldvar has been found in the formula, 
  // returns 0 otherwise (no substitution has been performed).
  unsigned int replaceOnce(char *oldvar,char *newvar);

  // returns the number of occurrences of the variable whose name is tosearch
  // in the current atom
  unsigned int occurrences(char *tosearch);

  // returns any positive integer if one or both the arithmetic
  // expressions composing the atom are mixed expressions
  // returns 0 otherwise
  unsigned int mixedAtom(void);

  // simply negates the atom (logical NOT)
  void negate(void);

  // returns a positive integer if the atom is identically true
  // 0 otherwise
  unsigned int isTrue(void);

  // returns a positive integer if the atom is identically false
  // 0 otherwise
  unsigned int isFalse(void);

  SI_operator getOp(void);

  // returns a string representation of the atom
  char* toString(void);
  char* toStringWithoutType(void);

  // returns a string representation of the atom in Sicstus clpfd syntax
  char* toStringInt(void);

  unsigned int operator > (SIatom p);

  unsigned int operator >= (SIatom p);

  unsigned int operator < (SIatom p);

  unsigned int operator <= (SIatom p);

  unsigned int operator == (SIatom p);

  // Returns the left expression composing the atom
  SIarithmetic leftExpr(void);

  // Return the right expression composing the atom
  SIarithmetic rightExpr(void);

};

#endif
