#ifndef SI_SIMPLIFIER
#define SI_SIMPLIFIER

#include "SIarithmetic.h"
#include "SIatom.h"
#include "SIformula.h"
#include "myvector.h"
#include "SicstusC.h"

// Class used for exceptions during simplifier initialization
// An exception of this class is raised when the simplifier initialization fails
// or when a SIsimplifier method is called when the SIsimplifier instance has not
// been initialized
// 
// Error codes: 
//
// 0: failure in SIsimplifier initialization
// 1: call to SIsimplifier method when the instance has not been initialized
class SIsimplerr {

 protected:

  unsigned int err;

 public:

  void setErr(unsigned int toset);

  unsigned int getErr();

};


class SIsimplifier {

 private:

  // Interface to the Sicstus process
  SicstusC iface;

  // Simplification support method
  void simplifyF(SIformula *tosimplify);

 protected:

  // monitor of successful initialization
  unsigned int initialized;

 public:
  
  // slot must be 0 or 1 (at most 2 prolog processes are allowed)
  // An exception is raised if the prolog initialization fails
  SIsimplifier(unsigned int slot);

  // Simplifies formula tosimplify.
  // Raises an exception if called on an uninitialized instance
  void simplify(SIformula *tosimplify);

  // returns -1 if simplifier forces  tosimplify to false
  // returns 0 if simplifier doesn't affect tosimplify (no subst.)
  // returns 1 if simplifier affects tosimplify but doesn't force it to false
  // returns 2 if simplifier is not a postcondition (nothing to  be done)
  // Raises an exception if called on an uninitialized instance
  int  propagate(SIformula *tosimplify,SIformula &simplifier);

  ~SIsimplifier(void);
};

#endif //SI_SIMPLIFIER
