#include "SIInterface.h"

#include "SIarithmetic.h"
#include <stdlib.h>
#include <stdio.h>


// Definition of comparison operator type
//enum SI_operator{gt,geq,lt,leq,eq,neq,unset};

SIInterface::SIInterface(void){
  this->state=UNINITIALIZED;
  this->SISimp = new SIsimplifier(0);
  this->replyCode[0]='\0';
}

const char * SIInterface::init(){
  if(this->state!=UNINITIALIZED){
    strcpy(this->replyCode, 
	   "554 Bad sequence. The simplifier is already initialized\n");
    this->reset();
    return this->replyCode;
  }
  this->state=NOFORMULAPRESENT;
  for(int i=0; i<ESTACKSIZE; i++){
    this->eStackForm[i]=NULL;
    this->eStackArit[i]=NULL;
  }
  this->formTop=0;
  this->aritTop=0;
  this->currentFormula=NULL;
  this->currentAssumptions=new SIformula(1);//currentAssumption=(true)
  strcpy(this->replyCode, "250 Requested simplifier action okay, completed\n");
  return this->replyCode;
}

const char * SIInterface::formulaStart(){
  if(this->state==UNINITIALIZED){
    strcpy(this->replyCode, 
	   "554 Bad sequence. Init the simplifier before sending formulae");
    this->reset();
    return this->replyCode;
  }
  if(this->state==FORMULAISPRESENT){
    strcpy(this->replyCode, 
	   "554 Bad sequence. Delete the old formula before sending new ones");
    this->reset();
    return this->replyCode;
  }
  if(this->state==BUILDINGAFORMULA){
    strcpy(this->replyCode, 
	   "554 Bad sequence. Already building a formula");
    this->reset();
    return this->replyCode;
  }
  this->state=BUILDINGAFORMULA;
  strcpy(this->replyCode, 
	 "354 Continue with formula input; end with the F0 command\n");
  return this->replyCode;
}

const char * SIInterface::formulaInteger(char *data){
  if(this->state!=BUILDINGAFORMULA){
    strcpy(this->replyCode, 
	   "554 Bad sequence. Enter the formula-building state first\n");
    this->reset();
    return this->replyCode;
  }
  if(this->aritTop>=ESTACKSIZE){
    strcpy(this->replyCode, 
	   "554 Transaction failed. Bigger evaluation stack is required\n");
    this->reset();
    return this->replyCode;
  }
  char *nonValidData;
  long intDatum = strtol(data,&nonValidData,10);
  if(strlen(data)==0 || strlen(nonValidData)!=0){
    strcpy(this->replyCode, 
	   "451 Requested action aborted: error in processing an int datum\n");
    return this->replyCode;
  }
  this->eStackArit[this->aritTop++]=new SIarithmetic(intDatum);
  strcpy(this->replyCode, 
	 "354 Continue with formula input; end with the F0 command\n");
  return this->replyCode;
}

const char * SIInterface::formulaReal(char *data){
  if(this->state!=BUILDINGAFORMULA){
    strcpy(this->replyCode, 
	   "554 Bad sequence. Enter the formula-building state first\n");
    this->reset();
    return this->replyCode;
  }
  if(this->aritTop>=ESTACKSIZE){
    strcpy(this->replyCode, 
	   "554 Transaction failed. Bigger evaluation stack is required\n");
    this->reset();
    return this->replyCode;
  }
  char *nonValidData;
  double doubleDatum=strtod((const char *)data,&nonValidData);
  if(strlen(data)==0 || strlen(nonValidData)!=0){
    strcpy(this->replyCode, 
	   "451 Requested action aborted: error in processing a real datum\n");
    return this->replyCode;
  }
  this->eStackArit[this->aritTop++]=new SIarithmetic((float)doubleDatum);
  strcpy(this->replyCode, 
	 "354 Continue with formula input; end with the F0 command\n");
  return this->replyCode;
}

const char * SIInterface::formulaBoolean(char *data){
  if(this->state!=BUILDINGAFORMULA){
    strcpy(this->replyCode, 
	   "554 Bad sequence. Enter the formula-building state first\n");
    this->reset();
    return this->replyCode;
  }
  if(this->formTop>=ESTACKSIZE){
    strcpy(this->replyCode, 
	   "554 Transaction failed. Bigger evaluation stack is required\n");
    this->reset();
    return this->replyCode;
  }
  unsigned int truthValue;
  if(strcmp(data,"true")==0)truthValue=1;
  else if (strcmp(data,"false")==0)truthValue=0;
  else{
    strcpy(this->replyCode, 
	   "451 Requested action aborted: error in processing a bool datum\n");
    return this->replyCode;
  }
  this->eStackForm[this->formTop++]=new SIformula(truthValue);
  strcpy(this->replyCode, 
	 "354 Continue with formula input; end with the F0 command\n");
  return this->replyCode;
}

const char * SIInterface::formulaSymbol(char *data){
  if(this->state!=BUILDINGAFORMULA){
    strcpy(this->replyCode, 
	   "554 Bad sequence. Enter the formula-building state first\n");
    this->reset();
    return this->replyCode;
  }
  if(this->aritTop>=ESTACKSIZE){
    strcpy(this->replyCode, 
	   "554 Transaction failed. Bigger evaluation stack is required\n");
    this->reset();
    return this->replyCode;
  }
  if(strlen(data)<2){
    strcpy(this->replyCode, 
	   "451 Requested action aborted: error in processing a symbol\n");
    return this->replyCode;
  }
  char type=data[0];
  char *name=data+1;
  if(type=='I')  
    this->eStackArit[this->aritTop++]=new SIarithmetic(name,SI_integer);
  else if(type=='R')  
    this->eStackArit[this->aritTop++]=new SIarithmetic(name,SI_real);
  else if(type=='B');//TBD
  else{
    strcpy(this->replyCode, 
	   "451 Requested action aborted: error in processing a symbol\n");
    return this->replyCode;
  }
  strcpy(this->replyCode, 
	 "354 Continue with formula input; end with the F0 command\n");
  return this->replyCode;
}

const char * SIInterface::formulaOperator(char *data){
  if(this->state!=BUILDINGAFORMULA){
    strcpy(this->replyCode, 
	   "554 Bad sequence. Enter the formula-building state first\n");
    this->reset();
    return this->replyCode;
  }
  if(strlen(data)!=1){
    strcpy(this->replyCode, 
	   "451 Requested action aborted: error in processing an operator\n");
    return this->replyCode;
  }

  char op=data[0];
  SIatom *atomicFormula;

  /* The construction of a formula based on an operator has been 
   * organized in 3 phases:
   * phase 1: check existence of operands and space in the evalution stack
   * phase 2: build the new formula
   * phase 3: clean operands and evaluation stack
   */
  switch(op){//construction of the new formula: phase 1
  case '+':
  case '-':
  case '*':
  case '/':
  case 'P':
  case 'E':
  case 'm':
  case 'M':
    if(this->aritTop<2){
      strcpy(this->replyCode, 
	     "554 Transaction failed. Arithmetic operator without operands\n");
      this->reset();
      return this->replyCode;
    }
    break;
  case '_':
  case '|':
  case 'S':
  case 'C':
  case 'T':
    if(this->aritTop<1){
      strcpy(this->replyCode, 
	     "554 Transaction failed. Arithmetic operator without operands\n");
      this->reset();
      return this->replyCode;
    }
    break;
  case '>':
  case 'G':
  case '=':
  case '!':
  case '<':
  case 'L':
    if(this->aritTop<2){
      strcpy(this->replyCode, 
	     "554 Transaction failed. Comparison operator without operands\n");
      this->reset();
      return this->replyCode;
    }
    else if(this->formTop>=ESTACKSIZE){
      strcpy(this->replyCode, 
	     "554 Transaction failed. Bigger evaluation stack is required\n");
      this->reset();
      return this->replyCode;
    }
    break;
  case 'A':
  case 'O':
  case 'I':
    if(this->formTop<2){
      strcpy(this->replyCode, 
	     "554 Transaction failed. Logic operator without operands\n");
      this->reset();
      return this->replyCode;
    }
    break;
  case 'N':
    if(this->formTop<1){
      strcpy(this->replyCode, 
	     "554 Transaction failed. Logic operator without operands\n");
      this->reset();
      return this->replyCode;
    }
    break;
  default:
    strcpy(this->replyCode, 
	   "451 Requested action aborted: error in processing an operator\n");
    return this->replyCode;
  }

  
  switch(op){//construction of the new formula: phase 2
  case '+':
    eStackArit[aritTop-2]->plus(*eStackArit[aritTop-1]);
    break;
  case '-':
    eStackArit[aritTop-2]->minus(*eStackArit[aritTop-1]);
    break;
  case '*':
    eStackArit[aritTop-2]->multiplied(*eStackArit[aritTop-1]);
    break;
  case '/':
    eStackArit[aritTop-2]->divided_by(*eStackArit[aritTop-1]);
    break;
  case 'P':
    eStackArit[aritTop-2]->power(*eStackArit[aritTop-1]);
    break;
  case 'E':
    eStackArit[aritTop-2]->power_exp(*eStackArit[aritTop-1]);
    break;
  case 'm':
    eStackArit[aritTop-2]->min(*eStackArit[aritTop-1]);
    break;
  case 'M':
    eStackArit[aritTop-2]->max(*eStackArit[aritTop-1]);
    break;
  case '_':
    eStackArit[aritTop-1]->negate();
    break;
  case '|':
    eStackArit[aritTop-1]->abs();
    break;
  case 'S':
    eStackArit[aritTop-1]->sin();
    break;
  case 'C':
    eStackArit[aritTop-1]->cos();
    break;
  case 'T':
    eStackArit[aritTop-1]->tan();
    break;
  case '>':
    atomicFormula=new SIatom(*eStackArit[aritTop-1],*eStackArit[aritTop-2],gt);
    break;
  case 'G':
    atomicFormula=new SIatom(*eStackArit[aritTop-1],*eStackArit[aritTop-2],geq);
    break;
  case '=':
    atomicFormula=new SIatom(*eStackArit[aritTop-1],*eStackArit[aritTop-2],eq);
    break;
  case '!':
    atomicFormula=new SIatom(*eStackArit[aritTop-1],*eStackArit[aritTop-2],neq);
    break;
  case '<':
    atomicFormula=new SIatom(*eStackArit[aritTop-1],*eStackArit[aritTop-2],lt);
    break;
  case 'L':
    atomicFormula=new SIatom(*eStackArit[aritTop-1],*eStackArit[aritTop-2],leq);
    break;
  case 'A':
    eStackForm[formTop-2]->SIand(*eStackForm[formTop-1]);
    break;
  case 'O':
    eStackForm[formTop-2]->SIor(*eStackForm[formTop-1]);
    break;
  case 'I':
    eStackForm[formTop-2]->SIimplies(*eStackForm[formTop-1]);
    break;
  case 'N':
    eStackForm[formTop-1]->SInot();
    break;
  }

  switch(op){//construction of the new formula: phase 3
  case '+':
  case '-':
  case '*':
  case '/':
  case 'P':
  case 'E':
  case 'm':
  case 'M':
    aritTop--;
    eStackArit[aritTop-1]->prts();
    delete eStackArit[aritTop];
    eStackArit[aritTop]=NULL;
    break;
  case '_':
  case '|':
  case 'S':
  case 'C':
  case 'T':
    eStackArit[aritTop-1]->prts();
    break;
  case '>':
  case 'G':
  case '=':
  case '!':
  case '<':
  case 'L':
    delete eStackArit[aritTop-1];
    delete eStackArit[aritTop-2];
    eStackArit[aritTop-1]=NULL;
    eStackArit[aritTop-2]=NULL;
    aritTop=aritTop-2;
    eStackForm[formTop++]=new SIformula(*atomicFormula);
    delete atomicFormula;
    break;
  case 'A':
  case 'O':
  case 'I':
    formTop--;
    delete eStackForm[formTop];
    eStackForm[formTop]=NULL;
    break;
  case 'N':
    break;
  }
  strcpy(this->replyCode, 
	 "354 Continue with formula input; end with the F0 command\n");
  return this->replyCode;
}

const char * SIInterface::formulaFinalize(){
  if(this->state!=BUILDINGAFORMULA){
    strcpy(this->replyCode, 
	   "554 Bad sequence. Enter the formula-building state first\n");
    this->reset();
    return this->replyCode;
  }
  if(aritTop!=0 || formTop!=1){
    strcpy(this->replyCode, 
	   "554 Transaction failed. Inconsistencies in the built formula\n");
    this->reset();
    return this->replyCode;
  }
  currentFormula=eStackForm[0];
  eStackForm[0]=NULL;
  formTop=0;
  this->state=FORMULAISPRESENT;
  strcpy(this->replyCode, "250 Requested simplifier action okay, completed\n");
  return this->replyCode;
}

const char * SIInterface::prove(bool proofGoal){
  if(this->state!=FORMULAISPRESENT){
    strcpy(this->replyCode, 
	   "554 Bad sequence. No formula ready for a proof");
    this->reset();
    return this->replyCode;
  }
  SIformula *formulaUnderEvaluation=new SIformula(*currentFormula);
  if(!proofGoal) formulaUnderEvaluation->SInot();
  SIformula *theorem=new SIformula(*currentAssumptions);
  theorem->SIimplies(*formulaUnderEvaluation);
  this->SISimp->simplify(theorem);
  bool proved = (theorem->isTrue()!=0);
  delete formulaUnderEvaluation;
  delete theorem;
  if(proved)
    strcpy(this->replyCode, 
	   "250 Requested simplifier action okay, completed\n");
  else 
    strcpy(this->replyCode, 
	   "451 Requested action aborted. Unable to prove the formula\n");
  return this->replyCode;
}

const char * SIInterface::append(bool logicAssumption){
  if(this->state!=FORMULAISPRESENT){
    strcpy(this->replyCode, 
	   "554 Bad sequence. No formula to append");
    this->reset();
    return this->replyCode;
  }
  if(!logicAssumption)currentFormula->SInot();
  currentAssumptions->SIand(*currentFormula);
  this->SISimp->simplify(currentAssumptions);
  delete currentFormula;
  currentFormula=NULL;
  this->state=NOFORMULAPRESENT;
  strcpy(this->replyCode, "250 Requested simplifier action okay, completed\n");
  return this->replyCode;
}

const char * SIInterface::discard(){
  if(this->state!=FORMULAISPRESENT){
    strcpy(this->replyCode, 
	   "554 Bad sequence. No formula to discard");
    this->reset();
    return this->replyCode;
  }
  delete currentFormula;
  currentFormula=NULL;
  this->state=NOFORMULAPRESENT;
  strcpy(this->replyCode, "250 Requested simplifier action okay, completed\n");
  return this->replyCode;
}

const char * SIInterface::clear(){
  for(int i=0; i<ESTACKSIZE; i++){ 
    if(this->eStackForm[i]) delete this->eStackForm[i];
    if(this->eStackArit[i]) delete this->eStackArit[i];
    this->eStackForm[i] = NULL;
    this->eStackArit[i] = NULL;
  }
  this->formTop=0;
  this->aritTop=0;

  if(this->currentFormula) delete this->currentFormula;
  this->currentFormula = NULL;
  if(this->currentAssumptions) delete this->currentAssumptions;
  this->currentAssumptions = new SIformula(1);//currentAssumption=(true)

  this->state=NOFORMULAPRESENT;

  strcpy(this->replyCode, "250 Requested simplifier action okay, completed\n");
  return this->replyCode;
}

const char * SIInterface::reset(){
  for(int i=0; i<ESTACKSIZE; i++){ 
    if(this->eStackForm[i])delete this->eStackForm[i];
    if(this->eStackArit[i])delete this->eStackArit[i];
  }
  if(this->currentAssumptions)delete this->currentAssumptions;
  if(this->currentFormula)delete this->currentFormula;
  this->state=UNINITIALIZED;
  strcpy(this->replyCode, "250 Requested simplifier action okay, completed\n");
  return this->replyCode;
}

SIInterface::~SIInterface(){
  for(int i=0; i<ESTACKSIZE; i++){ 
    if(this->eStackForm[i])delete this->eStackForm[i];
    if(this->eStackArit[i])delete this->eStackArit[i];
  }
  if(this->currentAssumptions)delete this->currentAssumptions;
  if(this->currentFormula)delete this->currentFormula;
  if(this->SISimp) delete this->SISimp;
}

