#include <string>
#include "myvector.h"
#include "utils.h"
#include "SIarithmetic.h"

// If debug mode is set, a trace of arithmetic constructors and destructors is kept

#ifdef TWO_DEBUG_MODE
int arithconstructors =0;
int arithdestructors =0;
#endif


/****************************** AUXILIARY METHODS ******************************/

// nonvarchars are not legal characters in variables names
unsigned int nonvarChar(char p) {
  return (p==' ') || (p=='+') || (p=='-') || (p=='*') || (p=='/') ||
    (p=='>') || (p=='<') || (p=='=') || (p=='\\') || (p==')') || (p=='(');
};

// controls whether the char is a digit or a dot
unsigned int numberChar(char a)
{
  return ( (a>='1') && (a<='9') ) || (a=='0') || (a=='.');
};

// a numberString is a string representing an integer number
unsigned int numberStr(char* str) {
  unsigned int isn=1;
  for(int cont=0;cont<strlen(str);cont++)
    { if (!numberChar(str[cont])) isn=0; }
  return isn;
};

// Tells whether the string totest is in toparse starting from
// the specified position
unsigned int contained(char* totest,std::string toparse,int position) {
  int totestLength = strlen(totest);
  int toparseLength = toparse.length();
  if(position >= toparseLength) return 0;
  if(totestLength == 0) return 0;
  if(position + totestLength > toparseLength) return 0;
  for(int cont = 0; cont < totestLength; cont++) {
    if(totest[cont] != toparse[position + cont]) return 0;
  }
  unsigned int endstr = (position + totestLength == toparseLength);
  unsigned int beginstr = (position == 0);
  if(!endstr)
    if(!nonvarChar(toparse[position + totestLength])) return 0;
  if(!beginstr)
    if(!nonvarChar(toparse[position-1])) return 0;
  return 1;
 }; 

// tells whether the input string represents a legal variable
// name
unsigned int legalVariable(char* totest) {
  std::string app;
  app=totest;
  if (app.empty()) return 0;
  if (numberChar(app[0])) return 0;
  return 1;
};

// tells whether the input string represents a pure number
unsigned int pureNumber(char* totest) {
  std::string app;
  app=totest;
  int dotcount=0;
  if (app.empty()) return 0;
  if (app[0]=='.') return 0;
  for(int cont=0;cont<app.length();cont++) {
    if (!numberChar(app[cont])) return 0;
    if (app[cont]=='.') dotcount++;
    if (dotcount>1) return 0;
  }
  return 1;
};

/************************** END OF AUXILIARY METHODS *****************************/



/******************* SIarithmetic private INTERFACE METHODS **********************/


// tells whether the variable is already in the list
unsigned int SIarithmetic::varIn(SIvariable t) {
  varlist.ResetCursor();
  for(int cont=0;cont<varlist.Length();cont++) {
    if(t==varlist.CursorElement()) return 1;
    varlist.IncCurs();
  }
  return 0;
};

// Support methods for arithmetic operators on strings

void SIarithmetic::plus(std::string toadd) {
  std::string app;
  app+="(";
  app+=expr;
  app+=")+(";
  app+=toadd;
  app+=")";
  expr=app;
};

void SIarithmetic::minus(std::string tosub) {
  std::string app;
  app+="(";
  app+=expr;
  app+=")-(";
  app+=tosub;
  app+=")";
  expr=app;
};

void SIarithmetic::multiplied(std::string tomult) {
  std::string app;
  app+="(";
  app+=expr;
  app+=")*(";
  app+=tomult;
  app+=")";
  expr=app;
};

void SIarithmetic::divided_by(std::string todiv) {
  std::string app;
  app+="(";
  app+=expr;
  app+=")/(";
  app+=todiv;
  app+=")";
  expr=app;
};

void SIarithmetic::power(std::string topow) {
  std::string app;
  app+="pow(";
  app+=expr;
  app+=", ";
  app+=topow;
  app+=")";
  expr=app;
};

void SIarithmetic::power_exp(std::string topow) {
  std::string app;
  app+="exp(";
  app+=expr;
  app+=", ";
  app+=topow;
  app+=")";
  expr=app;
};

void SIarithmetic::min(std::string tocompare) {
  std::string app;
  app+="min(";
  app+=expr;
  app+=", ";
  app+=tocompare;
  app+=")";
  expr=app;
};

void SIarithmetic::max(std::string tocompare) {
  std::string app;
  app+="max(";
  app+=expr;
  app+=", ";
  app+=tocompare;
  app+=")";
  expr=app;
};

// tells whether the expression is empty
unsigned int SIarithmetic::isEmpty(void) {
  return expr.length()==0;
};

/************** END OF SIarithmetic private INTERFACE METHODS ******************/ 




/************************* SIarithmeticErr methods ****************************/

void SIarithmeticErr::setCode(int toset) {
  errorcode=toset;
};

int SIarithmeticErr::getCode(void) {
  return errorcode;
};

/********************** END OF SIarithmeticErr methods **********************/





/**************************** SIvariable methods ***************************/


//constructor
SIvariable::SIvariable(void) {
  type=SI_unset;
};

// copy constructor
SIvariable::SIvariable(const SIvariable &p) {
  varname=p.varname;
  type=p.type;
};

// assignement operator
const SIvariable & SIvariable::operator = (const SIvariable &p) {
  varname=p.varname;
  type=p.type;
  return *this;
};

// sets the variable name to toset
void SIvariable::setVar(char *toset) {
  varname=toset;
};

// sets the variable type
void SIvariable::setType(SIarithtype toset) {
  type=toset;
};

// returns a string representation of the variable
char* SIvariable::getVar(void) {
  char* toreturn;
  toreturn=new char[varname.length()+1];
  strcpy(toreturn,varname.c_str());
  return toreturn;
};

// returns the variable type
SIarithtype SIvariable::getType(void) {
  return type;
};

// compares two variables
unsigned int SIvariable::operator==(SIvariable two){
  return varname==two.varname;
};

unsigned int SIvariable::operator<(SIvariable two) {
  return varname<two.varname;
};

unsigned int SIvariable::operator<=(SIvariable two) {
  return varname<=two.varname;
};

unsigned int SIvariable::operator>(SIvariable two) {
  return varname>two.varname;
};

unsigned int SIvariable::operator>=(SIvariable two) {
  return varname>=two.varname;
};

// The default destructor is enough
SIvariable::~SIvariable(void) {};


/*************************** END OF SIvariable methods **************************/




/********************* SIarithmetic public INTERFACE METHODS *******************/

SIarithmetic::SIarithmetic(void) {

#ifdef TWO_DEBUG_MODE
  arithconstructors++;
#endif

  exprtype=SI_unset;
  expr="";
};

//  OK, constructor validated by Sicolo
SIarithmetic::SIarithmetic(long a) {

#ifdef TWO_DEBUG_MODE
  arithconstructors++;
#endif

  exprtype=SI_integer;
  char* tmp_char=int_longToChar(a);
  expr=tmp_char;
  delete [] tmp_char;
};

// Builds an arithmetic expression from a float
SIarithmetic::SIarithmetic(float f) {

#ifdef TWO_DEBUG_MODE
  arithconstructors++;
#endif

  double a=f;
  char *app1=fchr(a);
  char *app2=mantissa(a);
  int one=strlen(app1);
  int two=strlen(app2);
  char *toset=new char[one+two+2];
  int cont;

  for(cont=0;cont<one;cont++) {
    toset[cont]=app1[cont];
  }
  toset[cont++]='.';
  for(int cont2=0;cont2<two;cont2++) {
    toset[cont++]=app2[cont2];
  }
  delete [] app1;
  delete [] app2;
  toset[cont++]='\0';
  expr=toset;
  delete [] toset;
  exprtype=SI_real;

};

// Creation of an arithmetic expression from a real number in rational
// format, e.g., A/B. Rational format is returned by Sicstus.
SIarithmetic::SIarithmetic(char* rational){
#ifdef TWO_DEBUG_MODE
  arithconstructors++;
#endif
  unsigned int formatok=1;
  unsigned int divcount=0;
  for(int i=0; i<strlen(rational); i++){
    switch(rational[i]){
    case '0': case '1': case '2': case '3': case '4':
    case '5': case '6': case '7': case '8': case '9':
      break;
    case '/': 
      divcount++; //non puo' esserci piu' di una '/'
      break;
    default: formatok=0;
    }
  }
  if(divcount!=1)formatok=0;
  if(!formatok) printf("\n\n\n***Warning: SIarithmetic::SIarithmetic: constructor's parameter is not in the expected format (i.e., A/B): %s\n\n\n",rational);
  expr = rational;
  exprtype = SI_real;
}

//  OK, constructor validated by Sicolo
SIarithmetic::SIarithmetic(char *varname, SIarithtype type) {

#ifdef TWO_DEBUG_MODE
  arithconstructors++;
#endif	  

  exprtype=SI_unset;
  this->setVar(varname,type);
};


//  OK, constructor validated by Sicolo
SIarithmetic::~SIarithmetic(void) {

#ifdef TWO_DEBUG_MODE
  arithdestructors++;
#endif

  //cout<<"####arithmetics "<<arithconstructors<<" "<<arithdestructors<<" "<<arithconstructors-arithdestructors<<endl;
};


// the type of variable (real or integer) must be provided along
// with varname. This method raises an exception if varname
// is not a legal variable, or if type is SI_unset
// A call to this method destroys the entire expression and just
// creates a new one constituted by variable varname
void SIarithmetic::setVar(char *varname, SIarithtype type) {

  SIarithmeticErr tothrow;
  std::string totest="";
  char* tmp_char;
  for(int cont=0;cont<strlen(varname);cont++) {
    totest+=varname[cont];
  }
  if (type==SI_unset) {
    tothrow.setCode(3);
    throw(tothrow);
  }
  tmp_char=stringToChar(totest);
  if (!legalVariable(tmp_char)) {
    delete [] tmp_char;
    tothrow.setCode(0);
    throw(tothrow);
  }
  delete [] tmp_char;
  expr=varname;
  exprtype=type;
  SIvariable toadd;
  toadd.setVar(varname);
  toadd.setType(type);
  while(varlist.Length()>0) varlist.Cut();
  varlist.Insert(toadd);
};

// Creates an expression constituted by the integer constant a
// A call to this method destroys the entire expression and just
// creates a new one constituted by the integer constant
void SIarithmetic::setLong(long a) {
  char* tmp_char=int_longToChar(a);
  expr=tmp_char;
  delete [] tmp_char;
  exprtype=SI_integer;
  while(varlist.Length()>0) varlist.Cut();
};

// Creates an expression constituted by the float constant a
// A call to this method destroys the entire expression and just
// creates a new one constituted by the float constant
void SIarithmetic::setDouble(double a) {
  char *app1=fchr(a);
  char *app2=mantissa(a);
  int one=strlen(app1);
  int two=strlen(app2);
  char *toset=new char[one+two+2];
  int cont;
  cont=0;
  for(;cont<one;cont++) {
    toset[cont]=app1[cont];
  }
  toset[cont++]='.';
  for(int cont2=0;cont2<two;cont2++) {
    toset[cont++]=app2[cont2];
  }
  delete [] app1;
  delete [] app2;
  toset[cont++]='\0';
  expr=toset;
  delete [] toset;
  exprtype=SI_real;
  while(varlist.Length()>0) varlist.Cut();
};


// sets the arithmetic expression to the passed string
// warning: this method can build invalid arithmetic expressions
// if not properly used
void SIarithmetic::setExpr(char *toset) {
  expr=toset;
  if (varlist.Length()==0) { 
    //se nessuna variabile  stata precedentemente aggiunta devo
    //occuparmi di settare il tipo dell'espressione (SI_integer o
    //SI_real)
    unsigned int number = 1;
    unsigned int divcount = 0;
    for(int i=0; i<strlen(toset); i++){
      switch(toset[i]){
      case '-':
	if(i!=0) number=0;
	break;
      case '0': case '1': case '2': case '3': case '4':
      case '5': case '6': case '7': case '8': case '9':
	break;
      case '/': 
	divcount++;
	break;
      default: number=0;
      }
    }
    if(divcount>1) number=0; //non puo' esserci piu' di una '/'    
    if(number && divcount==0) exprtype=SI_integer;
    else if(number && divcount==1) exprtype=SI_real;
    else{
      printf("\n\n\n***Warning: SIarithmetic::setExpr: expr does not represent a valid number: %s\n\n\n", toset);
      exprtype=SI_unset;
    }
  }  
};

// this method clears the variable list of the arithmetic expression
void SIarithmetic::clearVariables(void) {
  while(varlist.Length()>0) varlist.Cut();
};

// adds to the arithmetic expression the passed variable
// warning: this method can build invalid arithmetic expressions
// if not properly used
void SIarithmetic::addVariable(SIvariable toadd) {
  varlist.Insert(toadd);
  if(toadd.getType()==SI_real) 
    exprtype=SI_real;
  else if (toadd.getType()==SI_integer){
    if(exprtype!=SI_real) exprtype=SI_integer;
  }
  else 
    printf("\n\n\n***Warning: SIarithmetic::addVariable: Type of the variable added to the arithmetic expression is neither 'SI_integer' nor 'SI_real': The variable is %s\n\n\n",toadd.getVar());
};

// Adds p to the current arithmetic expression
void SIarithmetic::plus(SIarithmetic p) {

  if( (exprtype==SI_real) || (p.getType()==SI_real) ) exprtype=SI_real;
  else exprtype=SI_integer;
  
  plus(p.expr);
  
  // Variables list update
  p.varlist.ResetCursor();
  for(int cont=0;cont<p.varlist.Length();cont++) {
    if(!varIn(p.varlist.CursorElement())) varlist.Insert(p.varlist.CursorElement());
    p.varlist.IncCurs();
  }
  p.varlist.ResetCursor();
  
};

// Subtracts p to the current arithmetic expression
void SIarithmetic::minus(SIarithmetic p) {
  if( (exprtype==SI_real) || (p.getType()==SI_real) ) exprtype=SI_real;
  else exprtype=SI_integer;
  minus(p.expr);

  // Variables list update
  p.varlist.ResetCursor();
  for(int cont=0;cont<p.varlist.Length();cont++) {
    if(!varIn(p.varlist.CursorElement())) varlist.Insert(p.varlist.CursorElement());
    p.varlist.IncCurs();
  }
  p.varlist.ResetCursor();
};

// multiplies p to the current arithmetic expression
void SIarithmetic::multiplied(SIarithmetic p) {
  if( (exprtype==SI_real) || (p.getType()==SI_real) ) exprtype=SI_real;
  else exprtype=SI_integer;
  multiplied(p.expr);

  // Variables list update
  p.varlist.ResetCursor();
  for(int cont=0;cont<p.varlist.Length();cont++) {
    if(!varIn(p.varlist.CursorElement())) varlist.Insert(p.varlist.CursorElement());
    p.varlist.IncCurs();
  }
  p.varlist.ResetCursor();
};

// divides the current arithmetic expression by p
void SIarithmetic::divided_by(SIarithmetic p) {
  if( (exprtype==SI_real) || (p.getType()==SI_real) ) exprtype=SI_real;
  else exprtype=SI_integer;
  divided_by(p.expr);

  // Variables list update
  p.varlist.ResetCursor();
  for(int cont=0;cont<p.varlist.Length();cont++) {
    if(!varIn(p.varlist.CursorElement())) varlist.Insert(p.varlist.CursorElement());
    p.varlist.IncCurs();
  }
  p.varlist.ResetCursor();
};

// exponentiates the expression by p
void SIarithmetic::power(SIarithmetic p) {
  if( (exprtype==SI_real) || (p.getType()==SI_real) ) exprtype=SI_real;
  else exprtype=SI_integer;
  power(p.expr);

  // Variables list update
  p.varlist.ResetCursor();
  for(int cont=0;cont<p.varlist.Length();cont++) {
    if(!varIn(p.varlist.CursorElement())) varlist.Insert(p.varlist.CursorElement());
    p.varlist.IncCurs();
  }
  p.varlist.ResetCursor();
};

// exponentiates the expression by p
void SIarithmetic::power_exp(SIarithmetic p) {
  if( (exprtype==SI_real) || (p.getType()==SI_real) ) exprtype=SI_real;
  else exprtype=SI_integer;
  power_exp(p.expr);

  // Variables list update
  p.varlist.ResetCursor();
  for(int cont=0;cont<p.varlist.Length();cont++) {
    if(!varIn(p.varlist.CursorElement())) varlist.Insert(p.varlist.CursorElement());
    p.varlist.IncCurs();
  }
  p.varlist.ResetCursor();
};

// determines the minimum between the expression and p
void SIarithmetic::min(SIarithmetic p) {
  if( (exprtype==SI_real) || (p.getType()==SI_real) ) exprtype=SI_real;
  else exprtype=SI_integer;
  min(p.expr);

  // Variables list update
  p.varlist.ResetCursor();
  for(int cont=0;cont<p.varlist.Length();cont++) {
    if(!varIn(p.varlist.CursorElement())) varlist.Insert(p.varlist.CursorElement());
    p.varlist.IncCurs();
  }
  p.varlist.ResetCursor();
};

// determines the maximum between the expression and p
void SIarithmetic::max(SIarithmetic p) {
  if( (exprtype==SI_real) || (p.getType()==SI_real) ) exprtype=SI_real;
  else exprtype=SI_integer;
  max(p.expr);

  // Variables list update
  p.varlist.ResetCursor();
  for(int cont=0;cont<p.varlist.Length();cont++) {
    if(!varIn(p.varlist.CursorElement())) varlist.Insert(p.varlist.CursorElement());
    p.varlist.IncCurs();
  }
  p.varlist.ResetCursor();
};

// negates the current arithmetic expression
void SIarithmetic::negate()
{
  std::string app;
  app+="(0)-(";
  app+=expr;
  app+=")";
  expr=app;
};

void SIarithmetic::abs()
{
  std::string app;
  app+="abs(";
  app+=expr;
  app+=")";
  expr=app;
};

void SIarithmetic::sin()
{
  std::string app;
  app+="sin(";
  app+=expr;
  app+=")";
  expr=app;
};

void SIarithmetic::cos()
{
  std::string app;
  app+="cos(";
  app+=expr;
  app+=")";
  expr=app;
};

void SIarithmetic::tan()
{
  std::string app;
  app+="tan(";
  app+=expr;
  app+=")";
  expr=app;
};


unsigned int SIarithmetic::operator == (SIarithmetic p) {
  return expr==p.expr;
};


// returns the type of the arithmetic expression
SIarithtype SIarithmetic::getType() {
  return exprtype;
};

// returns all the variables contained in the arithmetic expression
// in a list
myvector<SIvariable> SIarithmetic::getVariables(void) {
  return varlist;
};

// returns any positive integer if the arithmetic expression is
// a mixed expression, i.e. if it contains both integer and
// real variables; returns 0 otherwise
unsigned int SIarithmetic::mixedExpr(void) {
  SIarithtype tp;
  if(varlist.Length()==0) return 0;
  tp=(varlist[0]).getType();
  varlist.ResetCursor();
  for(int cont=0;cont<varlist.Length();cont++) {
    if((varlist.CursorElement()).getType()!=tp) return 1;
    varlist.IncCurs();
  }
  return 0;
};


// replaces all occurrences of the variable whose name is in oldvar
// with the arithmetic expression newexpr
// returns a positive integer if at least an occurrence of the 
// variable whose name is in oldvar has been found in the formula, 
// returns 0 otherwise (no substitution has been performed).
unsigned int SIarithmetic::replace(char* oldvar,SIarithmetic newexpr) {
  std::string app;
  unsigned int substituted;
  char* tmp_char;
  app="";
  substituted=0;
  const char* curexpr=expr.c_str();
  
  // Visits the string representing the current expression searching for 
  // occurencies of oldvar; if found, they are replaced with newexpr
  // a temporary string (app) is used to build the new expression
  for(int cont=0;cont<strlen(curexpr);cont++) {
    if(contained(oldvar,curexpr,cont)) {
      substituted=1;
      tmp_char=newexpr.toStringWithoutType();
      app+=tmp_char;
      delete [] tmp_char;
      cont=cont+(strlen(oldvar)-1);
    }
    else {
      app+=curexpr[cont];
    }
  }
  
  tmp_char = stringToChar(app);
  expr=tmp_char;
  delete [] tmp_char;

  // Variables list update 1: searching for the replaced variable 
  if (substituted) {
    int pos=-1;
    varlist.ResetCursor();
    for(int cont2=0;cont2<varlist.Length();cont2++) {
      char* tmp=varlist.CursorElement().getVar();
      if (strcmp(tmp,oldvar)==0) {
	pos=varlist.CurPos();
      }
      varlist.IncCurs();
           
      delete [] tmp;
    }

    // Variables list update 2: eliminating the replaced variable
    if (pos>=0) varlist.Cut(pos);

    // Variables list update 3: inserting the variables of the replacing
    // arithmetic expression
    myvector<SIvariable> newlist=newexpr.getVariables();
    newlist.ResetCursor();
    for(int cont3=0;cont3<newlist.Length();cont3++) {
      if(!varIn(newlist.CursorElement())) varlist.Insert(newlist.CursorElement());
      newlist.IncCurs();
    }
  }

  return substituted;

};


// replaces all occurrences of the variable whose name is in oldvar
// with the variable name contained in newvar;
// if newvar is not a number, it is added to the current variables list with
// the same type as the replaced variable(oldvar)
// returns a positive integer if at least one occurrence of the 
// variable whose name is in oldvar has been found in the formula, 
// returns 0 otherwise (no substitution has been performed).
unsigned int SIarithmetic::replace(char* oldvar,char* newvar) {
  
  unsigned int isnumber=numberStr(newvar);
  std::string app;
  unsigned int substituted;
  char* tmp_char;
  app="";
  substituted=0;
  const char* curexpr=expr.c_str();

  // Visits the string representing the current expression searching for 
  // occurencies of oldvar; if found, they are replaced with newvar
  // a temporary string (app) is used to build the new expression
  for(int cont=0;cont<strlen(curexpr);cont++) {
    if(contained(oldvar,curexpr,cont)) {
      substituted=1;
      app+=newvar;
      cont=cont+(strlen(oldvar)-1);
    }
    else {
      app+=curexpr[cont];
    }
  }
  
  tmp_char = stringToChar(app);
  expr=tmp_char;
  delete [] tmp_char;

  unsigned int namein=0;

  // Controls whether newvar is already in the current variables list 
  varlist.ResetCursor();
  for(int parser=0;parser<varlist.Length();parser++) {
    std::string cmp1,cmp2;
    cmp1=newvar;
    char* oo=varlist.CursorElement().getVar();
    cmp2=oo;
    if(cmp1==cmp2) namein=1;
    varlist.IncCurs();
    delete [] oo;
  }
  
  // Variables list update 1: search for the substituted variable
  // tp stores the type of the removed variable
  // pos stores the variable position in the variables list
  if (substituted) {
    int pos=-1;
    SIarithtype tp;
    varlist.ResetCursor();
    for(int cont2=0;cont2<varlist.Length();cont2++) {
      char* tmp=varlist.CursorElement().getVar();
      if (strcmp(tmp,oldvar)==0) {
	pos=varlist.CurPos();
	tp=varlist.CursorElement().getType();
      }
      varlist.IncCurs();
      delete [] tmp;
    }

    // Variables list update 2: eliminates the substituted variable
    if (pos>=0) varlist.Cut(pos);

    // Variables list update 3: adds newvar to the variables list setting
    // its type to the type of the removed variable; the variables list is not
    // modified if a variable with the same name was already in varlist or if
    // newvar represents a number
    SIvariable toput;
    toput.setVar(newvar);
    toput.setType(tp);
    if( (!namein) && (!isnumber) ) varlist.Insert(toput);
  }

  return substituted;

};



// replaces all occurrences of the variable whose name is in oldvar
// with the variable name contained in newvar
// returns a positive integer if at least one occurrence of the 
// variable whose name is in oldvar has been found in the formula, 
// returns 0 otherwise (no substitution has been performed).
// This variant of the method doesn't add the new variable to
// variable list even though substituions have been performed 
unsigned int SIarithmetic::replaceNoAdd(char* oldvar,char* newvar) {

  std::string app;
  unsigned int substituted;
  char* tmp_char;
  app="";
  substituted=0;

  // Searches for oldvar in the current expression and substitutes it if found.
  // A temporary string (app) is used to build the new expression
  const char* curexpr=expr.c_str();
  for(int cont=0;cont<strlen(curexpr);cont++) {
    if(contained(oldvar,curexpr,cont)) {
      substituted=1;
      app+=newvar;
      cont=cont+(strlen(oldvar)-1);
    }
    else {
      app+=curexpr[cont];
    }
  }
  
  tmp_char=stringToChar(app);
  expr=tmp_char;
  delete [] tmp_char;
  unsigned int namein=0;

  // Controls whether newvar is already present in the current variables list
  varlist.ResetCursor();
  for(int parser=0;parser<varlist.Length();parser++) {
    std::string cmp1,cmp2;
    cmp1=newvar;
    char* oo=varlist.CursorElement().getVar();
    cmp2=oo;
    if(cmp1==cmp2) namein=1;
    varlist.IncCurs();
    delete [] oo;
  }

  // Variables list update: the replaced variable is searched and removed from 
  // the variables list
  if (substituted) {
    int pos=-1;
    SIarithtype tp;
    varlist.ResetCursor();
    for(int cont2=0;cont2<varlist.Length();cont2++) {
      char* tmp=varlist.CursorElement().getVar();
      if (strcmp(tmp,oldvar)==0) {
	pos=varlist.CurPos();
	tp=varlist.CursorElement().getType();
      }
      varlist.IncCurs();
      delete [] tmp;
    }
    if (pos>=0) varlist.Cut(pos);
  }

  return substituted;

};



// replaces only the first occurence of the variable whose name is in oldvar
// with the variable name contained in newvar;
// if newvar is not a number, it is added to the current variables list with
// the same type as the replaced variable(oldvar)
// returns a positive integer if at least one occurrence of the 
// variable whose name is in oldvar has been found in the formula, 
// returns 0 otherwise (no substitution has been performed).
unsigned int SIarithmetic::replaceOnce(char* oldvar,char* newvar) {
  
  unsigned int isnumber=numberStr(newvar);
  std::string app;
  unsigned int substituted,occurrences;
  char* tmp_char;
  app="";
  substituted=0;
  const char* curexpr=expr.c_str();

  // Visits the string representing the current expression searching for 
  // occurencies of oldvar; if found, they are replaced with newvar
  // a temporary string (app) is used to build the new expression
  occurrences=0;  // counts occurences of oldvar
  for(int cont=0;cont<strlen(curexpr);cont++) {
    if(contained(oldvar,curexpr,cont)) {
      if (occurrences<1) {
	substituted=1;
	app+=newvar;
	cont=cont+(strlen(oldvar)-1);
      }
      else { app+=curexpr[cont]; }
      occurrences++;
    }
    else {
      app+=curexpr[cont];
    }
  }
  
  tmp_char = stringToChar(app);
  expr=tmp_char;
  delete [] tmp_char;

  unsigned int namein=0;

  // Controls whether newvar is already in the current variables list 
  varlist.ResetCursor();
  for(int parser=0;parser<varlist.Length();parser++) {
    std::string cmp1,cmp2;
    cmp1=newvar;
    char* oo=varlist.CursorElement().getVar();
    cmp2=oo;
    if(cmp1==cmp2) namein=1;
    varlist.IncCurs();
    delete [] oo;
  }
  
  // Variables list update 1: search for the substituted variable
  // tp stores the type of the removed variable
  // pos stores the variable position in the variables list
  if (substituted) {
    int pos=-1;
    SIarithtype tp;
    varlist.ResetCursor();
    for(int cont2=0;cont2<varlist.Length();cont2++) {
      char* tmp=varlist.CursorElement().getVar();
      if (strcmp(tmp,oldvar)==0) {
	pos=varlist.CurPos();
	tp=varlist.CursorElement().getType();
      }
      varlist.IncCurs();
      delete [] tmp;
    }

    // Variables list update 2: eliminates the substituted variable, only if
    // no other occurence of the variable is left in the expression
    if ( (pos>=0) && (occurrences>1) ) varlist.Cut(pos);

    // Variables list update 3: adds newvar to the variables list setting
    // its type to the type of the removed variable; the variables list is not
    // modified if a variable with the same name was already in varlist or if
    // newvar represents a number
    SIvariable toput;
    toput.setVar(newvar);
    toput.setType(tp);
    if( (!namein) && (!isnumber) ) varlist.Insert(toput);
  }

  return substituted;

};


// counts the occurrences of tosearch in the current arithmetic expression
unsigned int SIarithmetic::occurrences(char* tosearch) {
  
  unsigned int occurrences;
  const char* curexpr=expr.c_str();

  // Visits the string representing the current expression searching for 
  // occurencies of oldvar; if found, they are replaced with newvar
  // a temporary string (app) is used to build the new expression
  occurrences=0;  // counts occurences of oldvar
  for(int cont=0;cont<strlen(curexpr);cont++) {
    if(contained(tosearch,curexpr,cont)) occurrences++;
  }

  return occurrences;

};

// returns a string representation of the arithmetic expression
char* SIarithmetic::toString(void) {
  SIarithmetic *toprint = new SIarithmetic(*this); 
  myvector<SIvariable> vars = toprint->getVariables();
  vars.ResetCursor();
  for(int i=0; i<vars.Length(); i++){
    char *name = vars.CursorElement().getVar();
    std::string typedname;
    if(vars.CursorElement().getType()==SI_real) 
      typedname = std::string("real::") + name;
    else if(vars.CursorElement().getType()==SI_integer) 
      typedname = std::string("int::") + name;
    else typedname = std::string("unset::") + name;
    char *newname = new char[typedname.length()+1];
    strcpy(newname, typedname.c_str());
    toprint->replace(name, newname);
    delete [] name;
    delete [] newname;
    vars.IncCurs();
  }
  std::string s=toprint->expr;
  char * toReturn=new char[s.length()+1];
  strcpy(toReturn,s.c_str());
  delete toprint;
  return toReturn;
};

char* SIarithmetic::toStringWithoutType(void) {

  std::string temp=expr;
  //SASSA: trucco per farlo digerire a sicstus
  if (expr[0]=='-')
    temp = std::string("0 ") + temp;
    
  char * toReturn=new char[temp.length()+1];
  strcpy(toReturn,temp.c_str());
  return toReturn;
};

// copy constructor
SIarithmetic::SIarithmetic(const SIarithmetic &p)
  :varlist(p.varlist), exprtype(p.exprtype), expr(p.expr){   // deep copy

#ifdef TWO_DEBUG_MODE
  arithconstructors++;
#endif
};

// assignement operator
const SIarithmetic & SIarithmetic::operator = (const SIarithmetic &p) {
  exprtype=p.exprtype;
  expr=p.expr;
  varlist=p.varlist;
  return *this;
};

// puts the arithmetic expression inside parenthesis
void SIarithmetic::prts(void) {
  std::string app;
  app+="(";
  app+=expr;
  app+=")";
  expr=app;
};

// tells wheter the expression is constituted by a single variable
unsigned int SIarithmetic::singleVariable(void) {
  if(varlist.Length()!=1) return 0;
  SIvariable one;
  one=varlist[0];
  std::string tt;
  char* oo=one.getVar();
  tt=oo;
  delete [] oo;  
  return tt==expr;
};

long SIarithmetic::getLValue() {

  // If an integer value of a non-numeric or of a real expression is requested
  // an exception is thrown
  if( (varlist.Length()>0) || (exprtype!=SI_integer) ) {
    SIarithmeticErr tothrow;
    tothrow.setCode(4);
    throw(tothrow);
  }

  long toreturn;
  char * tmp=new char[expr.length()+1];
  strcpy(tmp,expr.c_str());
  toreturn=stringToLong(tmp);
  delete [] tmp;
  return toreturn;
};

double SIarithmetic::getDValue() {

  // If a real value of a non-numeric or of an integer expression is requested
  // an exception is thrown
  if( (varlist.Length()>0) || (exprtype!=SI_real) ) {
    SIarithmeticErr tothrow;
    tothrow.setCode(5);
    throw(tothrow);
  }
  double toreturn;
  char * tmp=new char[expr.length()+1];
  strcpy(tmp,expr.c_str());
  toreturn=stringToDouble(tmp);
  delete [] tmp;
  return toreturn;
};

// Tells whether the expression is a pure number
unsigned int SIarithmetic::isNumber(void) {

  return ( (!isEmpty()) && (varlist.Length()==0) );

};

/**************** END OF SIarithmetic public INTERFACE METHODS *******************/
