#include "SIarithmetic.h"
#include "myvector.h"
#include "SIatom.h"
#include "utils.h"
#include <string>



/********************** SIatom private INTERFACE METHODS **************************/


// tells whether the variable is already in the list
unsigned int SIatom::varIn(SIvariable t) {

  varlist.ResetCursor();
  for(int cont=0;cont<varlist.Length();cont++) {
    if(t==varlist.CursorElement()) return 1;
    varlist.IncCurs();
  }
  return 0;

};

/******************* END OF SIatom private INTERFACE METHODS *********************/ 






/************************** SIatom INTERFACE METHODS ****************************/


SIatom::SIatom(void) {

  istrue=0;
  isfalse=0;
  op=unset;

};

SIatom::SIatom(SIarithmetic leftexp, SIarithmetic rightexp, SI_operator opr) {

  left=leftexp;
  right=rightexp;
  op=opr;

  // Creation of the atom variables list from the left and right expressions
  // variables lists
  while(varlist.Length()>0) { varlist.Cut(); }
  myvector<SIvariable> lf;
  lf=left.getVariables();
  lf.ResetCursor();
  for(int cont=0;cont<lf.Length();cont++) {
    varlist.Insert(lf.CursorElement());
    lf.IncCurs();
  }
  myvector<SIvariable> rg;
  rg=right.getVariables();
  rg.ResetCursor();
  for(int cont2=0;cont2<rg.Length();cont2++) {
    if(!varIn(rg.CursorElement())) varlist.Insert(rg.CursorElement());
    rg.IncCurs();
  }
  istrue=0;
  isfalse=0;

};
  
// A call to this method destroys the old atom an creates a new one
void SIatom::setAtom(SIarithmetic leftexp, SIarithmetic rightexp, SI_operator opr) {

  left=leftexp;
  right=rightexp;
  op=opr;
  while(varlist.Length()>0) { varlist.Cut(); }

  // Creation of the atom variables list starting from the left and right 
  // expressions variables list

  myvector<SIvariable> lf;
  lf=left.getVariables();
  lf.ResetCursor();
  for(int cont=0;cont<lf.Length();cont++) {
    varlist.Insert(lf.CursorElement());
    lf.IncCurs();
  }
  myvector<SIvariable> rg;
  rg=right.getVariables();
  rg.ResetCursor();
  for(int cont2=0;cont2<rg.Length();cont2++) {
    if(!varIn(rg.CursorElement())) varlist.Insert(rg.CursorElement());
    rg.IncCurs();
  }
  istrue=0;
  isfalse=0;

};

// reallineates the variables list according to the variables
// contained in the left and right arithmetic expressions
void SIatom::allineate(void) {

  while(varlist.Length()>0) { varlist.Cut(); }
  myvector<SIvariable> lf;
  lf=left.getVariables();
  lf.ResetCursor();
  for(int cont=0;cont<lf.Length();cont++) {
    varlist.Insert(lf.CursorElement());
    lf.IncCurs();
  }
  myvector<SIvariable> rg;
  rg=right.getVariables();
  rg.ResetCursor();
  for(int cont2=0;cont2<rg.Length();cont2++) {
    if(!varIn(rg.CursorElement())) varlist.Insert(rg.CursorElement());
    rg.IncCurs();
  }

};

// copy constructor
SIatom::SIatom(const SIatom& a) {

  left=a.left;
  right=a.right;
  op=a.op;
  istrue=a.istrue;
  isfalse=a.isfalse;
  varlist=a.varlist;

};

const SIatom & SIatom::operator = (const SIatom &a) {

  left=a.left;
  right=a.right;
  op=a.op;
  istrue=a.istrue;
  isfalse=a.isfalse;
  varlist=a.varlist;
  return *this;

};

// creates a FALSE atom if logicval==0, and a TRUE atom for all 
// other possible values of logicval
void SIatom::setLogicValue(unsigned int logicval) {

  SIarithmetic fake;
  setAtom(fake,fake,unset);
  while(varlist.Length()>0) { varlist.Cut(); }
  if (logicval==0) { isfalse=1; istrue=0; }
  else { isfalse=0; istrue=1; }
  
};

// returns all the variables contained in the atom in a list
myvector<SIvariable> SIatom::getVariables(void) {

  return varlist;

};

// replaces all occurrences of the variable whose name is in oldvar
// with the arithmetic expression newexpr
// returns a positive integer if at least an occurrence of the 
// variable whose name is in oldvar has been found in the formula, 
// returns 0 otherwise (no substitution has been performed).
unsigned int SIatom::replace(char *oldvar,SIarithmetic newexpr) {

  unsigned int subs1,subs2;
  subs1=subs2=0;
  subs1=left.replace(oldvar,newexpr);
  subs2=right.replace(oldvar,newexpr);
  allineate();
  return (subs1 || subs2);

};

// replaces all occurrences of the variable whose name is in oldvar
// with the variable name contained in newvar
// if newvar is not a number, it is added to the current variables list with
// the same type as the replaced variable(oldvar)
// returns a positive integer if at least an occurrence of the 
// variable whose name is in oldvar has been found in the formula, 
// returns 0 otherwise (no substitution has been performed).
unsigned int SIatom::replace(char *oldvar,char *newvar) {

  unsigned int subs1,subs2;
  subs1=subs2=0;
  subs1=left.replace(oldvar,newvar);
  subs2=right.replace(oldvar,newvar);
  allineate();
  return (subs1 || subs2);

};

// replaces only the first occurrence of the variable whose name is in oldvar
// with the variable name contained in newvar
// if newvar is not a number, it is added to the current variables list with
// the same type as the replaced variable(oldvar)
// returns a positive integer if at least an occurrence of the 
// variable whose name is in oldvar has been found in the formula, 
// returns 0 otherwise (no substitution has been performed).
unsigned int SIatom::replaceOnce(char *oldvar,char *newvar) {

  unsigned int subs1,subs2;
  subs1=subs2=0;
  subs1=left.replaceOnce(oldvar,newvar);
  if (!subs1) subs2=right.replaceOnce(oldvar,newvar);
  if (subs1 || subs2 ) allineate();
  return (subs1 || subs2);

};

// returns any positive integer if one or both the arithmetic
// expressions composing the atom are mixed expressions
// returns 0 otherwise
unsigned int SIatom::mixedAtom(void) {

  return (left.mixedExpr() || right.mixedExpr());

};

// simply negates the atom (logical NOT)
void SIatom::negate(void) {

  switch(op) {
  case gt:  op=leq;
    break;
  case geq: op=lt;
    break;
  case lt:  op=geq;
    break;
  case leq: op=gt;
    break;
  case eq:  op=neq;
    break;
  case neq: op=eq;
    break;
  default:  op=unset;
  };

};

// returns a positive integer if the atom is identically true
// 0 otherwise
unsigned int SIatom::isTrue(void) {

  return istrue;

};

// returns a positive integer if the atom is identically false
// 0 otherwise
unsigned int SIatom::isFalse(void){

  return isfalse;

};

SI_operator SIatom::getOp(void) {
  
  return op;

};

// returns a string representation of the atom
char* SIatom::toString(void) {

  std::string app;
  app+="( ";
  char* tmp=left.toString();
  app+=tmp;
  delete [] tmp;
  app+=" ";
  if(istrue) {
    std::string app2;
    app2+="true";
    return stringToChar(app2);
  }
  if(isfalse) {
    std::string app3;
    app3+="false";
    return stringToChar(app3);
  }
  switch(op) {
  case gt:  app+=">";
    break;
  case geq: app+=">=";
    break;
  case lt:  app+="<";
    break;
  case leq: app+="=<";
    break;
  case eq:  app+="=";
    break;
  case neq: app+="=";
    app+='\\';
    app+="=";
    break;
  default:  break;
  };
  app+=" ";
  char* tmp2=right.toString();
  app+=tmp2;
  delete [] tmp2;
  app+=" )";
  return stringToChar(app);

};

// returns a string representation of the atom but does not
// include type information for variables, numbers and symbols
char* SIatom::toStringWithoutType(void) {

  std::string app;
  app+="( ";
  char* tmp=left.toStringWithoutType();
  app+=tmp;
  delete [] tmp;
  app+=" ";
  if(istrue) {
    std::string app2;
    app2+="true";
    return stringToChar(app2);
  }
  if(isfalse) {
    std::string app3;
    app3+="false";
    return stringToChar(app3);
  }
  switch(op) {
  case gt:  app+=">";
    break;
  case geq: app+=">=";
    break;
  case lt:  app+="<";
    break;
  case leq: app+="=<";
    break;
  case eq:  app+="=";
    break;
  case neq: app+="=";
    app+='\\';
    app+="=";
    break;
  default:  break;
  };
  app+=" ";
  char* tmp2=right.toStringWithoutType();
  app+=tmp2;
  delete [] tmp2;
  app+=" )";
  return stringToChar(app);

};


// returns a string representation of the atom in Sicstus clpfd syntax
char* SIatom::toStringInt(void) {

  std::string app;
  app+="( ";
  char* tmp=left.toStringWithoutType();
  app+=tmp;
  delete [] tmp;
  app+=" ";
  if(istrue) {
    std::string app2;
    app2+="true";
    return stringToChar(app2);
  }
  if(isfalse) {
    std::string app3;
    app3+="false";
    return stringToChar(app3);
  }
  switch(op) {
  case gt:  app+="#>";
    break;
  case geq: app+="#>=";
    break;
  case lt:  app+="#<";
    break;
  case leq: app+="#=<";
    break;
  case eq:  app+="#=";
    break;
  case neq: app+="#";
    app+="\\";
    app+="=";
    break;
  default:  break;
  };
  app+=" ";
  char* tmp2=right.toStringWithoutType();
  app+=tmp2;
  delete [] tmp2;
  app+=" )";
  return stringToChar(app);

};

unsigned int SIatom::operator > (SIatom p) {

  std::string one,two;
  char* on=toString();
  one=on;
  char* tw=p.toString();
  two=tw;
  delete [] on;
  delete [] tw;
  return one>two;

};

unsigned int SIatom::operator >= (SIatom p) {

  std::string one,two;
  char* on=toString();
  one=on;
  char* tw=p.toString();
  two=tw;
  delete [] on;
  delete [] tw;
  return one>=two;

};

unsigned int SIatom::operator < (SIatom p) {
  
  std::string one,two;
  char* on=toString();
  one=on;
  char* tw=p.toString();
  two=tw;
  delete [] on;
  delete [] tw;
  return one<two;
  
};

unsigned int SIatom::operator <= (SIatom p) {
  
  std::string one,two;
  char* on=toString();
  one=on;
  char* tw=p.toString();
  two=tw;
  delete [] on;
  delete [] tw;
  return one<=two;

};

unsigned int SIatom::operator == (SIatom p) {

  std::string one,two;
  char* on=toString();
  one=on;
  char* tw=p.toString();
  two=tw;
  delete [] on;
  delete [] tw;
  return one==two;
  
};

unsigned int SIatom::occurrences(char *tosearch) {

  return left.occurrences(tosearch)+right.occurrences(tosearch);

};

SIarithmetic SIatom::leftExpr(void) {

  return left;

};

SIarithmetic SIatom::rightExpr(void) {

  return right;

};


/********************** END OF SIatom INTERFACE METHODS **************************/
