#include "SIatom.h"
#include "SIarithmetic.h"
#include "myvector.h"
#include "SIformula.h"
#include "utils.h"
#include <string>
#include <iostream>


// If two debug mode is set a trace of the constructed and destructed formulas is
// kept

#ifdef TWO_DEBUG_MODE
int constructor=0;
int SI_destructor=0;
#endif




/****************************** AUXILIARY METHODS ********************************/



// replaces oldvar with newvar in the entire  atomlist
// newvar is supposed to be belonging to the same type as oldvar
void replace(atomlist *t,char* oldvar,char* newvar) {

  t->resetCursor();
  int len=t->length();
  for(int cont=0;cont<len;cont++) {
    t->cursorElement().replace(oldvar,newvar);
    t->incCurs();
  }

};


// replaces the first occurrence of oldvar with newvar in the atomlist
// newvar is supposed to be belonging to the same type as oldvar
void replaceOnce(atomlist *t,char* oldvar,char* newvar) {

  t->resetCursor();
  int len=t->length();
  unsigned int replaced=0;
  for(int cont=0;cont<len;cont++) {
    if(!replaced)
      if(t->cursorElement().replaceOnce(oldvar,newvar)) replaced=1;
    t->incCurs();
  }

};

// replaces oldvar with newexpr in the entire  atomlist
void replace(atomlist *t,char* oldvar,SIarithmetic newexpr) {

  t->resetCursor();
  int len=t->length();
  for(int cont=0;cont<len;cont++) {
    t->cursorElement().replace(oldvar,newexpr);
    t->incCurs();
  }

};

// Tells whether the atom p is contained in atomlist k
unsigned int inside(SIatom p,atomlist k) {

  k.resetCursor();
  for(int cont=0;cont<k.length();cont++) {
    if (p==k.cursorElement()) return 1;
    k.incCurs();
  }
  return 0;

};

// tells whether the atomlist one is contained in atomlist two
unsigned int contained(atomlist one,atomlist two) {
  
  one.resetCursor();
  for(int cont=0;cont<one.length();cont++) {
    if(!inside(one.cursorElement(),two)) return 0;
    one.incCurs();
  }
  return 1;

};

// tells whether var is in tocontrol
unsigned int varIn(SIvariable var,myvector<SIvariable> tocontrol) {
  
  tocontrol.ResetCursor();
  for(int cont=0;cont<tocontrol.Length();cont++) {
    if(tocontrol.CursorElement()==var) return 1;
    tocontrol.IncCurs();
  };
  return 0;

};

/*************************** END OF AUXILIARY METHODS ****************************/




/************************** atomlist INTERFACE METHODS **************************/
 

// default constructor
atomlist::atomlist(void) {};

// copy constructor
atomlist::atomlist(atomlist const &p) {

  list=p.list;
  
};

// assignement operator
const atomlist & atomlist::operator = (const atomlist &p) {

  list=p.list;
  return *this;

};

// Cuts the last atom of the list
void atomlist::cut(void) {

  list.Cut();

};

// represents the atom list as a string
char* atomlist::writeList(void) {

  std::string app;
  app="";
  list.ResetCursor();
  for(int cont=0;cont<list.Length();cont++) {
    char* tmp=list.CursorElement().toString();
    app+=tmp;
    delete [] tmp;
    if(cont!=list.Length()-1) app+=" AND ";
    list.IncCurs();
  }
  return stringToChar(app);

};

// destructor
atomlist::~atomlist(void) {
  
  while(list.Length()>0) { list.Cut(); }
  
};

// Adds toadd to the list
void atomlist::add(SIatom toadd) {

  list.Insert(toadd);
  
};

// Resets the list the cursor to the first atom
void atomlist::resetCursor(void) {

  list.ResetCursor();
  
};

// Returns the atom currently pointed by cursor
SIatom& atomlist::cursorElement(void) {

  return list.CursorElement();

};

// Increases cursor by one position
void atomlist::incCurs(void) {

  list.IncCurs();

};

// Returns the variables list of the atom list 
myvector<SIvariable> atomlist::getVariables(void) {

  list.ResetCursor();
  myvector<SIvariable> toreturn;

  // for each element of the list, get the variables
  for(int cont=0;cont<list.Length();cont++) {
    myvector<SIvariable> cp=list.CursorElement().getVariables();
    cp.ResetCursor();
    // for each elemt of the got list, adds the variables in the variables vector
    // to be returned
    for(int cont2=0;cont2<cp.Length();cont2++) {
      if(!varIn(cp.CursorElement(),toreturn)) toreturn.Insert(cp.CursorElement());
      cp.IncCurs();
    }
    list.IncCurs();
  }
  return toreturn;

};

// Returns the number of elements of the list
int atomlist::length(void) {

  return list.Length();

};

// Set strict inclusion operator
unsigned int atomlist::operator < (atomlist p) {

  return contained(*this,p) && !(*this==p);

};

// Set general inclusion operator
unsigned int atomlist::operator <= (atomlist p) {

  return (*this<p) || (*this==p);

};

// Set strict contains operator
unsigned int atomlist::operator > (atomlist p) {

  return contained(p,*this) && !(*this==p);

};

// Set general contains operator
unsigned int atomlist::operator >= (atomlist p) {

  return (*this>p) || (*this==p);

};

// Set equality operator
unsigned int atomlist::operator == (atomlist p) {

  return (contained(*this,p)) && (contained(p,*this));

};

// concatenates two atom lists
void atomlist::concatenate(atomlist p) {

  p.list.ResetCursor();
  for(int cont=0;cont<p.list.Length();cont++) {
    list.Insert(p.list.CursorElement());
    p.list.IncCurs();
  }

};

unsigned int atomlist::occurrences(char *tosearch) {

  list.ResetCursor();
  int occ=0;
  for(int cont=0;cont<list.Length();cont++) {
    occ+=list.CursorElement().occurrences(tosearch);
    list.IncCurs();
  }
  return occ;

};


/*********************** END OF atomlist INTERFACE METHODS ***********************/




/************************** dnf_vector INTERFACE METHODS *************************/



dnf_vector::dnf_vector(void) {};
  
dnf_vector::dnf_vector(dnf_vector const &p) {

  vector=p.vector;

};

// assignement operator
const dnf_vector & dnf_vector::operator = (const dnf_vector &p) {

  vector=p.vector;
  return *this;

};

// returns the variables list of the dnf vector 
myvector<SIvariable> dnf_vector::getVariables() {

  myvector<SIvariable> toreturn;
  vector.ResetCursor();
  // for each element of the vector (atomlist) get the variables
  for(int cont=0;cont<vector.Length();cont++) {
    myvector<SIvariable> cp=vector.CursorElement().getVariables();
    cp.ResetCursor();
    // once the atomlist variables have been got, update the current variables list
    for(int cont2=0;cont2<cp.Length();cont2++) {
      if(!varIn(cp.CursorElement(),toreturn)) toreturn.Insert(cp.CursorElement());
      cp.IncCurs();
    }
    vector.IncCurs();
  }
  return toreturn;

};


dnf_vector::~dnf_vector(void) {
  
  while(vector.Length()>0) { vector.Cut(); }

};

// Transforms the dnf vector into a string
char* dnf_vector::writeVect(void) {

  std::string app;
  app="";
  vector.ResetCursor();
  for(int cont=0;cont<vector.Length();cont++) {  
    char* tmp=vector.CursorElement().writeList();
    app+=tmp;
    delete [] tmp;
    if(cont!=vector.Length()-1) app+="\n";
    vector.IncCurs();
  }
  return stringToChar(app);
};

// concatenates two dnf_vectors
dnf_vector dnf_vector::concatenate(dnf_vector one,dnf_vector two) {

  dnf_vector toreturn;
  one.resetCursor();
  two.resetCursor();
  int cont;
  for(cont=0;cont<one.length();cont++) {
    toreturn.add(one.cursorElement());
    one.incCurs();
  };
  for(cont=0;cont<two.length();cont++) {
    toreturn.add(two.cursorElement());
    two.incCurs();
  };
  return toreturn;

};

// Inserts toadd in the dnf vector
void dnf_vector::add(atomlist toadd) {

  vector.Insert(toadd);

};

// Reset the dnf vector cursor to position 0
void dnf_vector::resetCursor(void) {

  vector.ResetCursor();

};

// Erases the last atom list
void dnf_vector::cut(void) {
  
  vector.Cut();

};

// Erases the atom list at position i
void dnf_vector::cut(int i) {

  vector.Cut(i);

};

// Returns the current cursor position
int dnf_vector::curPos(void) {

  return vector.CurPos();

};

// index operator
atomlist dnf_vector::operator [] (const int index) {
  
  return vector[index];

};

// Inserts toadd in the dnf vector  
void dnf_vector::insert(atomlist toadd) {
  
  vector.Insert(toadd);

};

// returns the atom list currently pointed by cursor
atomlist& dnf_vector::cursorElement(void) {

  return vector.CursorElement();

}

// Increases cursor position by one
void dnf_vector::incCurs(void) {

  vector.IncCurs();

};

// Returns the number of atom lists composing the dnf vector
int dnf_vector::length(void) {

  return vector.Length();

};

// Erases the whole dnf vector
void dnf_vector::clean(void) {

  while(vector.Length()>0) { vector.Cut(); }

};

// concatenates the current dnf vector with p
dnf_vector dnf_vector::operator + (dnf_vector p) {

  dnf_vector toreturn;
  toreturn=concatenate(*this,p);
  return toreturn;

};

// algebraic product of current dnf vector with p
dnf_vector dnf_vector::operator * (dnf_vector p) {

  dnf_vector copy=*this;
  dnf_vector toreturn;
  copy.resetCursor();
  for(int cont=0;cont<copy.length();cont++) {
    p.resetCursor();
    for(int cont2=0;cont2<p.length();cont2++) {
      atomlist toadd;
      toadd.concatenate(copy.cursorElement());
      toadd.concatenate(p.cursorElement());
      toreturn.add(toadd);
      p.incCurs();
    }
    copy.incCurs();
  };
  return toreturn;
};


/************************* END OF dnf_vector INTERFACE METHODS *******************/




/*********************************** nodeErr METHODS *****************************/

void nodeErr::setCode(int toset) {

  errcode=toset;
  
};

int nodeErr::getCode(void) {

  return errcode;

};


/******************************** END OF nodeErr METHODS ************************/




/***************************** formulanode INTERFACE METHODS ********************/



// rebuilds the atom sequence on a tree
formulanode* formulanode::setNext(formulanode* next) {

  // the outer call of the method should have the NULL passed parameter

  // if the current node is an atom next is the next atom
  // and a pointer to the current node (atom) is returned
  if (this->isatom) {

    this->nextatom=next;
    return this;

  }

  // otherwise a recursive call should search for atom nodes
  else {
    formulanode* tempNext;
    tempNext=next;
    int cont;
    cont=this->children.Length()-1;
    for(;cont>=0;cont--) {
      tempNext=this->children[cont]->setNext(tempNext);
    }
    return(tempNext);
  }

};


formulanode::formulanode(void) {

  isatom=0;
  logicop=0;
  opr = SI_AND;
  nextatom=NULL;

};

// A call to this method sets the node to be an tom
// It raises an exception if one tries to set to atom a node that
// has already been defined
void formulanode::setAtom(SIatom toset) {

  // exception thrown when the state cannot accept the method call
  if(! ((isatom==0) && (logicop==0)) ) {
    nodeErr err;
    err.setCode(4);
    throw(err);
  };

  logicop=0;
  isatom=1;
  atm=toset;
  atomlist al;
  al.add(toset);
  dnform.add(al);
  nextatom=NULL;

};

// Sets the pointer to the next atom in the tree
void formulanode::setNextAtom(formulanode *toset) {

  nodeErr tothrow;
  if( (!isatom) && (!logicop) ) { tothrow.setCode(2); throw(tothrow); }
  else if(!isatom) { tothrow.setCode(1); throw(tothrow); }
  else if(!toset->isatom) { tothrow.setCode(3); throw(tothrow); }		
  { nextatom=toset; }
  
};

// A call to this method creates a logical operator with 
// the passed dnf vector
void formulanode::setLogicalOp(SI_logicalop op,dnf_vector dnf) {

  // An exception is thrown when the state does not allow the method call
  if(! ((isatom==0) && (logicop==0)) ) {
    nodeErr err;
    err.setCode(4);
    throw(err);
  };

  logicop=1;
  isatom=0;
  dnform=dnf;
  opr=op;

};

// adds a child to the tree
void formulanode::addChild(formulanode *toadd) {

  nodeErr tothrow;
  if( (!isatom) && (!logicop) ) { tothrow.setCode(2); throw(tothrow); }
  else if(!logicop) { tothrow.setCode(0); throw(tothrow); }
  else children.Insert(toadd);
  
};

// recursive destructor of the tree
formulanode::~formulanode(void) {

  if(this==NULL) return;
  else {
    while(dnform.length()>0) { dnform.cut(); }
    children.ResetCursor();
    for(int cont=0;cont<children.Length();cont++) {
      delete(children.CursorElement());
      children.IncCurs();
    }
    while(children.Length()>0) { children.Cut(); }
  }

};

// copy constructor: deep copy
// WARNING: this copy constructor doesn't replicate the atom
// sequence on the tree leaves
formulanode::formulanode(formulanode &x) {

  if( (x.isatom) || (x.logicop) ) {
      dnform=x.dnform;
      isatom=x.isatom;
      logicop=x.logicop;
      opr=SI_AND;
      // if the current node is an atom the copy is over
      // otherwise the copy must be recursively performed on children nodes
      if(x.isatom) { atm=x.atm; }
      else {
	opr=x.opr;
	myvector<formulanode*> tp=x.children;
	tp.ResetCursor();
	for(int cont=0;cont<tp.Length();cont++) {
	  addChild(new formulanode(*(tp.CursorElement())));
	  tp.IncCurs();
	}
      }
  }
  else {
    isatom=x.isatom;
    logicop=x.logicop;
    opr=x.opr;
    nextatom=NULL;
  }

};

// adjusts all the dnf of the nodes of the tree
// replacing oldvar with newvar
void formulanode::adjust(char* oldvar,char* newvar) {

  dnform.resetCursor();
  for(int cont=0;cont<dnform.length();cont++) {
    replace(&dnform.cursorElement(),oldvar,newvar);
    dnform.incCurs();
  }
  children.ResetCursor();
  for(int cont=0;cont<children.Length();cont++) {
    (children.CursorElement())->adjust(oldvar,newvar);
    children.IncCurs();
  }

};

// adjusts all the dnf of the nodes of the tree
// replacing oldvar with newexpr
void formulanode::adjust(char* oldvar,SIarithmetic newexpr) {

  dnform.resetCursor();
  for(int cont=0;cont<dnform.length();cont++) {
    replace(&dnform.cursorElement(),oldvar,newexpr);
    dnform.incCurs();
  }
  children.ResetCursor();
  for(int cont=0;cont<children.Length();cont++) {
    (children.CursorElement())->adjust(oldvar,newexpr);
    children.IncCurs();
  }

};


/*************************** END OF formulanode METHODS **************************/





/********************** SIformula private INTERFACE METHODS **********************/



// Indicates the last atom of the formula (last leaf)
formulanode* SIformula::lastAtom(formulanode* p) {

  if (p->isatom) return p;
  else return lastAtom(p->children[p->children.Length()-1]);

};

// Indicates the first atom of the formula (first leaf)
formulanode* SIformula::firstAtom(formulanode* p) {

  if (p->isatom) return p;
  else return firstAtom(p->children[0]);

};


// creates a string representation of the subtree starting from tw
char* SIformula::writeNode(formulanode *tw) {

  std::string app;

  // if tw is an atom, simply gets the string representation of it
  if (tw->isatom) {
    char* tmp=tw->atm.toString();
    app=tmp;
    delete [] tmp;
    return stringToChar(app);
  }

  // otherwise writes the subformulas pointed by the children vector with the
  // correct logical operator parenthesis. A recursive call is performed on
  // the children nodes
  else {
    int cont;

    // NOT operator
    if (tw->opr==SI_NOT) {
      app+="NOT ( ";
      myvector<formulanode*> tp=tw->children;
      tp.ResetCursor();
      for(cont=0;cont<tp.Length();cont++) {
	char* tmp2=writeNode(tp.CursorElement());
	app+=tmp2;
	delete [] tmp2;
	tp.IncCurs();
      }
      app+=" )";
    }

    // AND operator
    if (tw->opr==SI_AND) {
      app+="( ";
      myvector<formulanode*> tp=tw->children;
      tp.ResetCursor();
      for(cont=0;cont<tp.Length();cont++) {
	char* tmp3=writeNode(tp.CursorElement());
	app+=tmp3;
	delete [] tmp3;
	if(tp.CurPos()!=(tp.Length()-1)) app+=" AND ";
	tp.IncCurs();
      }
      app+=" )";
    }

    // OR operator
    if (tw->opr==SI_OR) {
      app+="( ";
      myvector<formulanode*> tp=tw->children;
      tp.ResetCursor();
      for(cont=0;cont<tp.Length();cont++) {
	char* tmp4=writeNode(tp.CursorElement());
	app+=tmp4;
	delete [] tmp4;
	if(tp.CurPos()!=(tp.Length()-1)) app+=" OR ";
	tp.IncCurs();
      }
      app+=" )";
    }
    return stringToChar(app);
  }

};


/****************** END OF SInformula private INTERFACE METHODS *******************/





/********************* SInformula public INTERFACE METHODS ************************/


// default constructor
SIformula::SIformula(void) {

#ifdef TWO_DEBUG_MODE
  constructor++;
  //cout<<"-----costruisco nformula "<<constructor<<endl;
#endif
  istrue=isfalse=0;
  top=NULL;
  firstatom=NULL;
  lastatom=NULL;

};


// OK, constructor validated  by Sicolo
// Creates a FALSE formula if logicval=0, creates a TRUE formula otherwise
SIformula::SIformula(unsigned int logicval) {

#ifdef TWO_DEBUG_MODE
  constructor++;
  //cout<<"-----costruisco nformula "<<constructor<<endl;
#endif
  top=NULL;
  firstatom=NULL;
  lastatom=NULL;
  if(logicval==0) { istrue=0; isfalse=1; }
  else { istrue=1; isfalse=0; }

};


// Transform the atom p into a formula
SIformula::SIformula(SIatom p) {

#ifdef TWO_DEBUG_MODE
  constructor++;
  //cout<<"-----costruisco nformula "<<constructor<<endl;
#endif
  formulanode* n=new formulanode();
  istrue=isfalse=0;
  n->setAtom(p);
  firstatom=n;
  lastatom=n;
  top=n;

};

// creates a formula starting from the passed atom
// a call to this method destroys the instance to create a new one
void SIformula::setAtom(SIatom p) {

  delete(top);
  formulanode* n=new formulanode();
  n->setAtom(p);
  firstatom=n;
  lastatom=n;
  top=n;

};

// creates a FALSE formula if logicval==0
// creates a TRUE formula for all other possible values of logicval
// a call to this method destroys the instance to create a new one
void SIformula::setLogicVal(unsigned int logicval) {

  delete(top);
  top=firstatom=lastatom=NULL;
  if(logicval==0) { istrue=0; isfalse=1; }
  else { istrue=1; isfalse=0; }

};

// copy constructor (deep copy)
SIformula::SIformula(const SIformula &p) {

#ifdef TWO_DEBUG_MODE
  constructor++;
  //cout<<"-----costruisco formula "<<constructor<<endl;
#endif

  // IF p.top is not null a recursive call is needed, unless the formula to be
  // copied is identically TRUE or FALSE
  if(p.top!=NULL) {
    istrue=p.istrue;
    isfalse=p.isfalse;
    if ( (!istrue) && (!isfalse) ) {
      top=new formulanode(*(p.top));
      firstatom=firstAtom(top);
      lastatom=lastAtom(top);
    }
    else {
      top=firstatom=lastatom=NULL;
    }

    // Re-constructs the atom chain in the copied formula
    if (top!=NULL) {
      (*top).setNext(NULL);
    }
  }

  // if p.top==NULL a simple copy of the node is needed
  else {
    istrue=p.istrue;
    isfalse=p.isfalse;
    top=firstatom=lastatom=NULL;
  }

};

// Assignement operator
const SIformula & SIformula::operator =(const SIformula & p) {

  // Works in the same way as the copy constructor, but destroys the current
  // object before building the copy of p
  if( (p.istrue) || (p.isfalse) ) {
    istrue=p.istrue;
    isfalse=p.isfalse;
    delete top;  // current object destruction
    top=firstatom=lastatom=NULL;
  }
  else {
    istrue=p.istrue;
    isfalse=p.isfalse;
    delete top; // current object destruction
    top=new formulanode(*(p.top));
    // Atoms chain re-construction
    (*top).setNext(NULL);
    firstatom=firstAtom(top);
    lastatom=lastAtom(top);
  }
  
  return *this;
  
};


// Creates a new formula: current formula = current formula AND p
void SIformula::SIand(SIformula p) {

  // current AND TRUE = current
  if(p.isTrue()) {
    return;
  }

  // current AND FALSE = FALSE, FALSE AND p = FALSE
  if(p.isFalse() || this->isFalse()) {
    delete top;
    firstatom=lastatom=top=NULL;
    istrue=0;
    isfalse=1;
    return;
  }

  // TRUE AND p = p
  if (this->isTrue()) {
    delete top;
    top=new formulanode(*p.top);
    firstatom=this->firstAtom();
    lastatom=this->lastAtom();
    istrue=p.isTrue();
    isfalse=p.isFalse();
    top->setNext(NULL);
    return;
  }

  // normal case: neither p nor current are TRUE or FALSE
  if(p.top!=NULL) {

    // If the current formula is already an AND conjunction, simply adds a child
    if ( (top->logicop) && (top->opr==SI_AND) ) {
      formulanode *toadd=new formulanode(*(p.top));
      top->children.Insert(toadd);
      // Atoms chain re-construction
      lastatom->setNextAtom(firstAtom(toadd));
      lastatom=lastAtom(toadd);
      // dnform update (algebraic product)
      top->dnform=(top->dnform)*(p.top->dnform);
    }
    // if the current formula is not an AND conjunction, creates an AND node with
    // children pointing to current and p 
    else {
      formulanode *np=new formulanode();
      np->setLogicalOp(SI_AND,(top->dnform)*(p.top->dnform));
      np->children.Insert(top);
      formulanode *toadd=new formulanode(*(p.top));
      np->addChild(toadd);
      // atoms chain re-construction
      lastatom->setNextAtom(firstAtom(toadd));
      lastatom=lastAtom(toadd);
      top=np;
    }
    //update atoms chain
    top->setNext(NULL);

  };

};

// Creates a new formula: current formula = current formula OR p
void SIformula::SIor(SIformula p) {

  // current OR TRUE = TRUE; TRUE OR p = TRUE
  if(p.isTrue()||this->isTrue()) {
    delete top;
    firstatom=lastatom=top=NULL;
    istrue=1;
    isfalse=0;
    return;
  }

  // current OR FALSE = current
  if(p.isFalse()) {
    return;
  }

  // FALSE OR p = p
  if (this->isFalse()) {
    
    delete top;
    top=new formulanode(*(p.top));
    
    firstatom=firstAtom(top);
    lastatom=lastAtom(top);
    istrue=p.isTrue();
    isfalse=p.isFalse();
    top->setNext(NULL);
    return;
  }
  
  // normal case: neither p nor current are TRUE or FALSE
  if(p.top!=NULL) {

    // If the current formula is an OR addition, simply adds p as a child
    if ( (top->logicop) && (top->opr==SI_OR) ) {

      formulanode *toadd=new formulanode(*(p.top));
      top->children.Insert(toadd);
      lastatom->setNextAtom(firstAtom(toadd));
      lastatom=lastAtom(toadd);
      top->dnform=(top->dnform)+(p.top->dnform);

    }

    // If the current formula is not an OR addition, creates an OR node with
    // current and p as children
    else {
      formulanode *np=new formulanode();
      np->setLogicalOp(SI_OR,(top->dnform)+(p.top->dnform));
      np->children.Insert(top);
      formulanode *toadd=new formulanode(*(p.top));
      np->children.Insert(toadd);
      lastatom->setNextAtom(firstAtom(toadd));
      lastatom=lastAtom(toadd);
      top=np;
    }
    //update atoms chain
    top->setNext(NULL);
  };

};

// Creates a new formula: NOT(current)
void SIformula::SInot() {

  // NOT TRUE = FALSE
  if(istrue) {
    istrue=0;
    isfalse=1;
    return;
  }

  // NOT FALSE = TRUE
  if(isfalse) {
    istrue=1;
    isfalse=0;
    return;
  }

  // normal case : the current formula is neither TRUE nor FALSE
  if(top!=NULL) {

    // if the current formula is an atom, simply negates the atom
    if(top->isatom) { 
      top->atm.negate();
      // dnform update
      while(top->dnform.length()>0) top->dnform.cut();
      SIatom nwatm=top->atm;
      atomlist nwlist;
      nwlist.add(nwatm);
      top->dnform.insert(nwlist);
      return; 
    }

    // if the current formula is already a logical negation, simply eliminates
    // the top NOT node. The dnform of the node pointed by children[0] is already
    // updated to be the new formula dnform and needs no update
    if(top->opr==SI_NOT) {
      formulanode *todel=top;
      top=top->children[0];
      todel->children.Cut();
      delete(todel);
      return;
    }

    // if the current formula is not a negation, several steps must be performed

    // dnform update procedure
    //
    // 1. negates all the atom lists of the current dnf_vector
    //    1.a negates all the atoms of the atom list
    //    1.b creates a new dnf_vector containing all the negated atoms
    // 2. put the negated atom lists in an AND conjunction (algebraic product)
    dnf_vector toset;
    dnf_vector tp=top->dnform;
    int cont,cont2;
    tp.resetCursor();
    for(cont=0;cont<tp.length();cont++) {
      dnf_vector dnftomult;
      atomlist atl=tp.cursorElement();
      atl.resetCursor();
      for(cont2=0;cont2<atl.length();cont2++) {
	atomlist fake;
	SIatom nn;
	nn=atl.cursorElement();
	nn.negate();
	fake.add(nn);
	dnftomult.add(fake);
	atl.incCurs();
      }
      if(cont==0) toset=dnftomult;
      else toset=toset*dnftomult;
      tp.incCurs();
    }
    formulanode *nw=new formulanode();
    nw->setLogicalOp(SI_NOT,toset);
    nw->children.Insert(top);
    top=nw;
  };

};

// creates a new formula: current = current IMPLIES p
void SIformula::SIimplies(SIformula p) {

  SInot();
  SIor(p);

};

// creates a string representation of the current formula using writeNode  
char* SIformula::writeFormula(void) {

  std::string app;
  if (istrue) { app+="true"; return stringToChar(app); }
  else if (isfalse) { app+="false"; return stringToChar(app); }
  else if(top==NULL) {
    char *nullreturn;
    nullreturn=new char[1];
    nullreturn[0]='\0'; 
    return nullreturn; }
  else return writeNode(top);

};

// replaces all occurrences of the variable whose name is in oldvar
// with the arithmetic expression newexpr
// returns a positive integer if at least one occurrence of the 
// variable whose name is in oldvar has been found in the formula, 
// returns 0 otherwise (no substitution has been performed).
unsigned int SIformula::replace(char *oldvar,SIarithmetic newexpr) {

  formulanode* app;
  unsigned int substituted;
  substituted=0;
  app=firstatom;
  while(app!=NULL) {
    if(app->atm.replace(oldvar,newexpr)) substituted=1;
    app=app->nextatom;
  }
  // updates the dnform of all nodes according to the performed substitution
  if ( (!istrue) && (!isfalse) && (substituted) )
    top->adjust(oldvar,newexpr);
  return substituted;

};

// replaces all occurrences of the variable whose name is in oldvar
// with the variable name contained in newvar
// returns a positive integer if at least one occurrence of the 
// variable whose name is in oldvar has been found in the formula, 
// returns 0 otherwise (no substitution has been performed).
unsigned int SIformula::replace(char *oldvar,char *newvar) {

  formulanode* app;
  unsigned int substituted;
  substituted=0;
  app=firstatom;
  while(app!=NULL) {
    if(app->atm.replace(oldvar,newvar)) substituted=1;
    app=app->nextatom;
  }
  if ( (!istrue) && (!isfalse) && (substituted) )  
    top->adjust(oldvar,newvar);
  return substituted;

};

// destructor
SIformula::~SIformula(void) {
#ifdef TWO_DEBUG_MODE
  SI_destructor++;
  std::cout << "FORMULA DESTRUCTOR: #constructed: " << constructor << ", #destructed: " << SI_destructor << ", difference: " << constructor-SI_destructor << std::endl;
  // std::cout << "Chiamata al metodo distruttore di SIformula.\n";
#endif
  
  firstatom=lastatom=NULL;
  istrue=isfalse=0;
  delete(top);
  top=NULL;
   
};

// returns 0 if the instance is not an AND conjunction of equality 
// atoms each with a single variable on the left, 
// any other positive value if it is
unsigned int SIformula::eqConj(void) {
  
  if(top==NULL) return 0;
  if ( (!top->isatom) && (!top->logicop) ) return 0;
  if ( top->isatom ) {
    if ( top->atm.getOp()!=eq ) return 0;
    // if the left expression is not a single variable the formula is not
    // a required equality conjunction
    if ( !top->atm.leftExpr().singleVariable() ) return 0;
    return 1;
  }
  if (top->opr!=SI_AND) return 0;
  myvector<formulanode *> tolook=top->children;
  tolook.ResetCursor();
  for(int cont=0;cont<tolook.Length();cont++) {
    formulanode *tp;
    tp=tolook.CursorElement();
    if( (tp->logicop ) || ( (!tp->logicop) && (!tp->isatom) ) ) return 0;
    if ( (tp->atm).getOp()!=eq ) return 0;
    if ( ! (tp->atm).leftExpr().singleVariable() ) return 0;
    tolook.IncCurs();
  }
  return 1;
};

// returns all the variables contained in the formula in a list
myvector<SIvariable> SIformula::getVariables(void){

  myvector<SIvariable> toreturn;
  formulanode *app;
  app=firstatom;
  while(app!=NULL) {
    myvector<SIvariable> cnt;
    cnt=(app->atm).getVariables();
    cnt.ResetCursor();
    for(int cont=0;cont<cnt.Length();cont++){
      if(!varIn(cnt.CursorElement(),toreturn))  
	toreturn.Insert(cnt.CursorElement()); 
      cnt.IncCurs();
    }
    app=app->nextatom;
  }
  return toreturn;

};



// returns the pairs variable-value contained in the formula
// precondition: the formula is a well-formed postcondition
std::map<std::string,SIarithmetic*> SIformula::getVarValues(){

  std::map<std::string,SIarithmetic*> result;
  formulanode *app;
  app=firstatom;
  while(app!=NULL) {
    char* tmp_charstr = app->atm.leftExpr().toStringWithoutType();
    result[tmp_charstr]=new SIarithmetic(app->atm.rightExpr());
    delete[] tmp_charstr;
    app=app->nextatom;
  }
  return result;
}

//returns the atoms contained in the formula
//no information is given about the structure of the formula
std::vector<SIatom*> SIformula::getAtoms(){

  std::vector<SIatom*> result;
  formulanode *app;
  app = firstatom;
  while(app!=NULL){
    result.push_back(new SIatom(app->atm));
    app = app->nextatom;
  }
  return result;
}



// returns any positive integer if the formula contains mixed 
// atoms; returns 0 otherwise
unsigned int SIformula::mixedFormula(void) {

  formulanode* app;
  app=firstatom;
  while(app!=NULL)
    { if ((app->atm).mixedAtom())  return 1; app=app->nextatom; }
  return 0;
  
};

unsigned int SIformula::isFalse(void) {

  return isfalse;

};

unsigned int SIformula::isTrue(void) {

  return istrue;

};

formulanode* SIformula::firstAtom(void) {

  return firstatom;

};

formulanode* SIformula::lastAtom(void) {

  return lastatom;

};

dnf_vector SIformula::dnForm(void) {

  if (top!=NULL) return top->dnform;
  else {  
    dnf_vector fake;
    return  fake;
  }

};



/**************** END OF SInformula public INTERFACE METHODS **********************/
