#include "SIarithmetic.h"
#include "SIatom.h"
#include "SIformula.h"
#include "myvector.h"
#include "utils.h"
#include "SicstusC.h"
#include "SIsimplifier.h"
#include <iostream>
#include <string>
using namespace std;


// The prolog interface
SicstusC *prologchannel;



/******************************* SUPPORT CLASSES **********************************/



//////////////////////////////////// SICSTUS NAME /////////////////////////////////

// class used to generate and store sicstus compliant variable names

class sicstus_name {

protected: 

  string cnt;

public: 

  sicstus_name(void) {

    cnt="A";

  };
  
  void reset(void) {

    cnt="A";

  };

  char* toChar(void) {

    return stringToChar(cnt);

  };

  void increase(void) {

    switch(cnt[cnt.length()-1]) {
    case 'A' : cnt[cnt.length()-1]='B';
      break;
    case 'B': cnt[cnt.length()-1]='C';
      break;
    case 'C': cnt[cnt.length()-1]='D';
      break;
    case 'D': cnt[cnt.length()-1]='E';
      break;      
    case 'E': cnt[cnt.length()-1]='F';
      break;
    case 'F': cnt[cnt.length()-1]='G';
      break;
    case 'G': cnt[cnt.length()-1]='H';
      break;
    case 'H': cnt[cnt.length()-1]='I';
      break;
    case 'I': cnt[cnt.length()-1]='J';
      break;
    case 'J': cnt[cnt.length()-1]='K';
      break;
    case 'K': cnt[cnt.length()-1]='L';
      break;
    case 'L': cnt[cnt.length()-1]='M';
      break;
    case 'M': cnt[cnt.length()-1]='N';
      break;
    case 'N': cnt[cnt.length()-1]='O';
      break;
    case 'O': cnt[cnt.length()-1]='P';
      break;
    case 'P': cnt[cnt.length()-1]='Q';
      break;
    case 'Q': cnt[cnt.length()-1]='R';
      break;
    case 'R': cnt[cnt.length()-1]='S';
      break;
    case 'S': cnt[cnt.length()-1]='T';
      break;
    case 'T': cnt[cnt.length()-1]='U';
      break;
    case 'U': cnt[cnt.length()-1]='V';
      break;
    case 'V': cnt[cnt.length()-1]='W';
      break;
    case 'W': cnt[cnt.length()-1]='X';
      break;
    case 'X': cnt[cnt.length()-1]='Y';
      break;
    case 'Y': cnt[cnt.length()-1]='Z';
      break;
    case 'Z': cnt+='A';
      break;
    default:
      break;
    }
  };

};

///////////////////////////////////////////////////////////////////////////////////



//////////////////////////////////// MININAME /////////////////////////////////////

// class used to generate and store variables names used in the premangling phase

class mininame {

protected:
 
  string cnt;

public: 

  mininame(void) {

    cnt="a";

  };
  
  void reset(void) {

    cnt="a";

  };

  char* toChar(void) {

    return stringToChar(cnt);

  };

  void increase(void) {

    switch(cnt[cnt.length()-1]) {
    case 'a' : cnt[cnt.length()-1]='b';
      break;
    case 'b': cnt[cnt.length()-1]='c';
      break;
    case 'c': cnt[cnt.length()-1]='d';
      break;
    case 'd': cnt[cnt.length()-1]='e';
      break;      
    case 'e': cnt[cnt.length()-1]='f';
      break;
    case 'f': cnt[cnt.length()-1]='g';
      break;
    case 'g': cnt[cnt.length()-1]='h';
      break;
    case 'h': cnt[cnt.length()-1]='i';
      break;
    case 'i': cnt[cnt.length()-1]='j';
      break;
    case 'j': cnt[cnt.length()-1]='k';
      break;
    case 'k': cnt[cnt.length()-1]='l';
      break;
    case 'l': cnt[cnt.length()-1]='m';
      break;
    case 'm': cnt[cnt.length()-1]='n';
      break;
    case 'n': cnt[cnt.length()-1]='o';
      break;
    case 'o': cnt[cnt.length()-1]='p';
      break;
    case 'p': cnt[cnt.length()-1]='q';
      break;
    case 'q': cnt[cnt.length()-1]='r';
      break;
    case 'r': cnt[cnt.length()-1]='s';
      break;
    case 's': cnt[cnt.length()-1]='t';
      break;
    case 't': cnt[cnt.length()-1]='u';
      break;
    case 'u': cnt[cnt.length()-1]='v';
      break;
    case 'v': cnt[cnt.length()-1]='w';
      break;
    case 'w': cnt[cnt.length()-1]='x';
      break;
    case 'x': cnt[cnt.length()-1]='y';
      break;
    case 'y': cnt[cnt.length()-1]='z';
      break;
    case 'z': cnt+='a';
      break;
    default:
      break;
    }
  };

};

///////////////////////////////////////////////////////////////////////////////////



/////////////////////////////////////// NAME COUPLE //////////////////////////////

// Class used to store pairs of variable names

class name_couple {

protected: 

  char *one;
  char *two;

public:

  name_couple() {

    one=NULL;
    two=NULL;

  };
  
  name_couple(const name_couple& s) {

    one=new char[strlen(s.one)+1];
    strcpy(one,s.one);
    two=new char[strlen(s.two)+1];
    strcpy(two,s.two);
    
  };
  
  void setOne(char* ts) {

    string cp;
    cp=ts;
    delete [] one;
    one=stringToChar(cp);
    
  };
  
  void setTwo(char* ts) {

    string cp;
    cp=ts;
    delete [] two;
    two=stringToChar(cp);

  };
  
  char* getOne(void) {

    char* toreturn;
    toreturn= new char[strlen(one)+1];
    strcpy(toreturn,one);
    return toreturn;

  };
  
  char* getTwo(void) {

    char* toreturn;
    toreturn= new char[strlen(two)+1];
    strcpy(toreturn,two);
    return toreturn;

  };
  
  const name_couple & operator =(const name_couple& n) {

    delete [] one;
    delete [] two;
    if (n.one) {
      one=new char[strlen(n.one)+1];
      strcpy(one,n.one);
    }
    else one=NULL;
    if (n.two) {
      two=new char[strlen(n.two)+1];
      strcpy(two,n.two);
    }else two=NULL;
    
    return *this;

  }
  
  ~name_couple(void) {

    delete [] one;
    one=NULL;
    delete [] two; 
    two=NULL;

  };

};

///////////////////////////////////////////////////////////////////////////////////



//////////////////////////////////// VAREXPR COUPLE //////////////////////////////


// Class used to store variable-arithmetic expression pairs

class varexpr_couple {

protected:
 
  char *one;
  SIarithmetic two;

public:

  varexpr_couple() {
    
    one=NULL;
    
  };
  
  varexpr_couple(const varexpr_couple& s) {

    one=new char[strlen(s.one)+1];
    strcpy(one,s.one);
    two=s.two;

  };
  
  void setOne(char* ts) {

    string cp;
    cp=ts;
    delete [] one;
    one=stringToChar(cp);

  };
  
  void setTwo(SIarithmetic ts) {

    two=ts;

  };
  
  char* getOne(void) {

    char* toreturn;
    toreturn= new char[strlen(one)+1];
    strcpy(toreturn,one);
    return toreturn;

  };
  
  SIarithmetic getTwo(void) {

    return two;

  };
  
  const varexpr_couple & operator =(const varexpr_couple& n) {
    
    delete [] one;
    if (n.one) {
      one=new char[strlen(n.one)+1];
      strcpy(one,n.one);
    }
    else one=NULL;
    
    two=n.two;
    return *this;
    
  };
  
  ~varexpr_couple(void) {
    
    delete [] one; one=NULL;
    
  };
  
};

///////////////////////////////////////////////////////////////////////////////////



/**************************** END OF SUPPORT CLASSES ******************************/




/******************************* SUPPORT METHODS *********************************/


// Returns 0 if totest contains no variables, 1 if totest contains only integer
// variables, 2 if totest contains only real variables, 3 if totest contains
// both integer and real variables
unsigned int typeRecognizeAtomlist(atomlist totest) {
  myvector<SIvariable> vars=totest.getVariables();
  if (vars.Length()==0) return 0;
  unsigned int allinteger,allreal;
  allinteger=allreal=1;
  vars.ResetCursor();
  for(int cont=0;cont<vars.Length();cont++) {
    SIvariable vr=vars.CursorElement();
    if(vr.getType()==SI_real) allinteger=0;
    if(vr.getType()==SI_integer) allreal=0;
    vars.IncCurs();
  }
  if (allinteger) return 1;
  if (allreal) return 2;
  return 3;

};

unsigned int typeRecognizeAtom(SIatom totest) {
  myvector<SIvariable> vars=totest.getVariables();
  if (vars.Length()==0) return 0;
  unsigned int allinteger,allreal;
  allinteger=allreal=1;
  vars.ResetCursor();
  for(int cont=0;cont<vars.Length();cont++) {
    SIvariable vr=vars.CursorElement();
    if(vr.getType()==SI_real) allinteger=0;
    if(vr.getType()==SI_integer) allreal=0;
    vars.IncCurs();
  }
  if (allinteger) return 1;
  if (allreal) return 2;
  return 3;

};

char* getIntVarsToCommaSeparatedString(atomlist totest){
  myvector<SIvariable> vars = totest.getVariables();
  if (vars.Length() == 0) return NULL;
  string s;
  vars.ResetCursor();
  for(int cont = 0; cont < vars.Length(); cont++){
    SIvariable vr = vars.CursorElement();
    if(vr.getType() == SI_integer){
      char* vname = vr.getVar();
      s += vname;
      delete vname;
    }
    vars.IncCurs();
  }
  char * ret = new char[s.length()+1];
  strcpy(ret, s.c_str());
  return ret;
};


// Tells whether two is the logical negation of one
unsigned int atomNegation(SIatom one,SIatom two) {
  
  unsigned int corr,inv;
  corr=(one.leftExpr()==two.leftExpr()) && (one.rightExpr()==two.rightExpr());
  inv=(one.leftExpr()==two.rightExpr()) && (one.rightExpr()==two.leftExpr());
  if ( (!corr) && (!inv) ) return 0;
  SI_operator ll,rr;
  ll=one.getOp();
  rr=two.getOp();
  if( (corr) && (ll==gt) ) if (rr!=leq) return 0;
  if( (corr) && (ll==geq) ) if (rr!=lt) return 0;
  if( (corr) && (ll==lt) ) if (rr!=geq) return 0;
  if( (corr) && (ll==leq) ) if (rr!=gt) return 0;
  if( (corr) && (ll==eq) ) if (rr!=neq) return 0;
  if( (corr) && (ll==neq) ) if (rr!=eq) return 0;
  
  if( (inv) && (ll==gt) ) if (rr!=geq) return 0;
  if( (inv) && (ll==geq) ) if (rr!=gt) return 0;
  if( (inv) && (ll==lt) ) if (rr!=leq) return 0;
  if( (inv) && (ll==leq) ) if (rr!=lt) return 0;
  if( (inv) && (ll==eq) ) if (rr!=neq) return 0;
  if( (inv) && (ll==neq) ) if (rr!=eq) return 0;
  
  return 1;
  
};

// Tells whether two is the logical negation of one
unsigned int listNegation(atomlist one,atomlist two) {

  if (one.length()!=1) return 0;
  if (two.length()!=1) return 0;
  one.resetCursor();
  two.resetCursor();
  return atomNegation(one.cursorElement(),two.cursorElement());

};

unsigned int isNumber(char p) {

  return ( ( (p>='1') && (p<='9') )  || (p=='0') );

};

unsigned int isLetter(char p) {

  return ( (p>='A') && (p<='Z') );

};

// Tells whetehr t is a reserved operator character
unsigned int oprChar(char t) {

  return ( (t=='=') || (t=='\\') || (t=='>') || (t=='<') );

};

// Tells whether p is an uppercase letter
unsigned int isUpperCaseLetter(char p) {

  switch(p) {
  case 'A': return 1;
    break;
  case 'B': return 1;
    break;
  case 'C': return 1;
    break;
  case 'D': return 1;
    break;
  case 'E': return 1;
    break;
  case 'F': return 1;
    break;
  case 'G': return 1;
    break;
  case 'H': return 1;
    break;
  case 'I': return 1;
    break;
  case 'J': return 1;
    break;
  case 'K': return 1;
    break;
  case 'L': return 1;
    break;
  case 'M': return 1;
    break;
  case 'N': return 1;
    break;
  case 'O': return 1;
    break;
  case 'P': return 1;
    break;
  case 'Q': return 1;
    break;
  case 'R': return 1;
    break;
  case 'S': return 1;
    break;
  case 'T': return 1;
    break;
  case 'U': return 1;
    break;
  case 'V': return 1;
    break;
  case 'X': return 1;
    break;
  case 'Y': return 1;
    break;
  case 'Z': return 1;
    break;
  default: return 0;
  }

};

// Tells whether toparse is made up of uppercase letters only
unsigned int allUpperCase(char *toparse) {

  int cont=strlen(toparse);
  if (cont==0) return 0;
  for(int parse=0;parse<cont;parse++) {
    if (!isUpperCaseLetter(toparse[parse])) return 0;
  }
  return 1;

};

// Returns the variables occurring in t
myvector<SIvariable> vrs(atomlist t) {

  myvector<SIvariable> toreturn;
  t.resetCursor();
  for(int cont=0;cont<t.length();cont++) {
    myvector<SIvariable> cp=t.cursorElement().getVariables();
    cp.ResetCursor();
    for(int cont2=0;cont2<cp.Length();cont2++) {
      if(!varIn(cp.CursorElement(),toreturn)) toreturn.Insert(cp.CursorElement());
      cp.IncCurs();
    }
    t.incCurs();
  }
  return toreturn;

};

// eliminates all the variables names made up of uppercase letters only
// in the atomlist t
myvector<name_couple> presymbolTable(atomlist t) {

  t.resetCursor();
  mininame crp;
  crp.reset();
  myvector<SIvariable> vars=vrs(t);
  vars.ResetCursor();
  myvector<name_couple> toreturn;

  for(int cont=0;cont<vars.Length();cont++) {
    name_couple nc;
    char* tmp=vars.CursorElement().getVar();
    // Only uppercase-only variables are mangled in the premangle phase
    if(allUpperCase(tmp)) {
      nc.setOne(tmp);
      delete [] tmp;
      tmp=crp.toChar();
      nc.setTwo(tmp);
      delete [] tmp;
      crp.increase();
      toreturn.Insert(nc);
    }
    else delete [] tmp;
    vars.IncCurs();
  }
  return toreturn;

};


// mangles the names of all the  variables of the atomlist t
myvector<name_couple> symbolTable(atomlist t) {

  t.resetCursor();
  sicstus_name crp;
  crp.reset();
  myvector<SIvariable> vars=vrs(t);
  vars.ResetCursor();
  myvector<name_couple> toreturn;
  for(int cont=0;cont<vars.Length();cont++) {
    name_couple nc;
    char* tmp=vars.CursorElement().getVar();
    nc.setOne(tmp);
    delete [] tmp;
    tmp=crp.toChar();
    nc.setTwo(tmp);
    delete [] tmp;
    crp.increase();
    vars.IncCurs();
    toreturn.Insert(nc);
  }
  return toreturn;

};


// mangles the names of all the  variables of the mixed atomlist t
// mx returns the number of entries in the symbol table that refer to
// multi-mangled real variables (each occurrence of a real variable is
// mangled with a different name)
myvector<name_couple> symbolTableForMixed(atomlist t,unsigned int *mx) {

  t.resetCursor();
  sicstus_name crp;
  crp.reset();
  myvector<SIvariable> vars=vrs(t);

  myvector<SIvariable> reals; // used to store only the real variables of vars
  myvector<SIvariable> integers; // used to store only the integer variables of vars

  // counts the entries in the table that refer to real variables mangling
  unsigned int mxmng=0;

  // cycles counter
  int cont;

  vars.ResetCursor();

  // separate real variables from integer variables
  for(cont=0;cont<vars.Length();cont++) {
    if (vars.CursorElement().getType()==SI_real)
      reals.Insert(vars.CursorElement());
    else
      integers.Insert(vars.CursorElement());
    vars.IncCurs();
  }

  myvector<name_couple> toreturn;

  // prepares the mangle table for real variables
  reals.ResetCursor();
  for(cont=0;cont<reals.Length();cont++) {

    int thisoccur;
    thisoccur=t.occurrences(reals.CursorElement().getVar());
    mxmng+=thisoccur;
    for(int cont2=0;cont2<thisoccur;cont2++) {
      name_couple nc;
      char* tmp=reals.CursorElement().getVar();
      nc.setOne(tmp);
      delete [] tmp;
      tmp=crp.toChar();
      nc.setTwo(tmp);
      delete [] tmp;
      crp.increase();
      toreturn.Insert(nc);
    }
    reals.IncCurs();

  }

  integers.ResetCursor();
  for(cont=0;cont<integers.Length();cont++) {
    name_couple nc;
    char* tmp=integers.CursorElement().getVar();
    nc.setOne(tmp);
    delete [] tmp;
    tmp=crp.toChar();
    nc.setTwo(tmp);
    delete [] tmp;
    crp.increase();
    integers.IncCurs();
    toreturn.Insert(nc);
  }

  *mx=mxmng;
  return toreturn;

};


//  performs the substitutions dictated by subs in the  atomlist
void mangle(atomlist *t,myvector<name_couple> subs) {

  subs.ResetCursor();
  for(int cont=0;cont<subs.Length();cont++) {
    char *one,*two;
    one=subs.CursorElement().getOne();
    two=subs.CursorElement().getTwo();
    replace(t,one,two);
    delete [] one; one=NULL;
    delete [] two; two=NULL;
    subs.IncCurs();
  }

};


// performs the substitutions dictated by subs in the  atomlist,
// using multiple mangling for real variables, indicated by the rnumber
// (that is the number of entries in subs that refer to real variables)
void mangleForMixed(atomlist *t,myvector<name_couple> subs,unsigned int rnumber) {

  subs.ResetCursor();
  // cycle counter
  int cont;

  // mangling of real variables
  for(cont=0;cont<rnumber;cont++) {
    char *one,*two;
    one=subs.CursorElement().getOne();
    two=subs.CursorElement().getTwo();
    replaceOnce(t,one,two);
    delete [] one; one=NULL;
    delete [] two; two=NULL;
    subs.IncCurs();
  }

  // mangling of integer variables
  for(cont=rnumber;cont<subs.Length();cont++) {
    char *one,*two;
    one=subs.CursorElement().getOne();
    two=subs.CursorElement().getTwo();
    replace(t,one,two);
    delete [] one; one=NULL;
    delete [] two; two=NULL;
    subs.IncCurs();
  }

};


//  performs the substitutions dictated by subs in the  atomlist
void premangle(atomlist *t,myvector<name_couple> subs) {

  subs.ResetCursor();
  for(int cont=0;cont<subs.Length();cont++) {
    char *one,*two;
    one=subs.CursorElement().getOne();
    two=subs.CursorElement().getTwo();
    replace(t,one,two);
    delete [] one; one=NULL;
    delete [] two; two=NULL;
    subs.IncCurs();
  }

};


//  performs the inverse substitutions dictated by subs in the  atomlist
void demangle(atomlist *t,myvector<name_couple> subs) {

  subs.ResetCursor();
  for(int cont=0;cont<subs.Length();cont++) {
    char *one,*two;
    one=subs.CursorElement().getOne();
    two=subs.CursorElement().getTwo();
    replace(t,two,one);
    delete [] one; one=NULL;
    delete [] two; two=NULL;
    subs.IncCurs();
  }

};


//  performs the inverse substitutions dictated by subs in the  atomlist
void depremangle(atomlist *t,myvector<name_couple> subs) {

  subs.ResetCursor();
  for(int cont=0;cont<subs.Length();cont++) {
    char *one,*two;
    one=subs.CursorElement().getOne();
    two=subs.CursorElement().getTwo();
    replace(t,two,one);
    delete [] one; one=NULL;
    delete [] two; two=NULL;
    subs.IncCurs();
  }

};

// Prepares a Sicstus clpr query string representing the atomlist t
char* realQuery(atomlist t) {

  string tmp;
  
#ifdef TWO_DEBUG_MODE
  cout<< "Atom list prima di preparare il comando.";
  t.resetCursor();
  for(int cont2=0;cont2<t.length();cont2++) {
    char* tmp2=(t.cursorElement().toStringWithoutType());
    cout << tmp2 << "\n";
    t.incCurs();
    delete [] tmp2;
  }
#endif

  t.resetCursor();
  
  for(int cont=0;cont<t.length();cont++) {
    tmp+="{";
    char* tmp3=t.cursorElement().toStringWithoutType();
    tmp+=tmp3;
    tmp+="}";
    delete [] tmp3;
    if(cont < t.length()-1) tmp+=",";
    t.incCurs();
  }
  tmp+=".";

#ifdef TWO_DEBUG_MODE
  cout << "Realquery ha preparato il comando: " << tmp << "\n"; 
#endif
  return stringToChar(tmp);

};

// Prepares a Sicstus clpfd query string representing the atomlist t
char* integerQuery(atomlist t) {
#ifdef TWO_DEBUG_MODE

  if(t.length()>0) 
    cout << "integerQuery: atomlist length >0.\n";
  else cout << "integerQuery: atomlist length =0!!!\n"; 
  cout<< "integerequery: atomlist: \n";
  t.resetCursor();
  for(int cont3=0;cont3<t.length();cont3++) {
    char* tmp=t.cursorElement().toStringWithoutType();
    cout<< tmp << "\n";
    delete [] tmp;
    t.incCurs();
  }
  
#endif


  string tmp;
  t.resetCursor();
  for(int cont=0;cont<t.length();cont++) {
    tmp+="(";
    char* hh=t.cursorElement().toStringInt();
#ifdef TWO_DEBUG_MODE
    cout << "integerquery ha riconosciuto i caratteri: " << hh << "\n"; 
#endif
    tmp+=hh;
    tmp+=")";
    delete [] hh;
    if(cont<t.length()-1) tmp+=",";
    t.incCurs();
  }
  tmp+=".";

#ifdef TWO_DEBUG_MODE
  cout << "integerquery ha preparato il comando: " << tmp << "\n"; 
#endif

  return stringToChar(tmp);

};

// tells whether the string contains '='
unsigned int equalIn(char* totest) {

  for(int cont=0;cont<strlen(totest);cont++) {
    if (totest[cont]=='=') return 1;
  }
  return 0;

};

//  if returned value is >0, one and two contain the two expressions of the
//  equality atom totest
unsigned int divideEq(char* totest,char** one,char** two) {

  if(!equalIn(totest)) return 0;
  string h1,h2;
  unsigned int found=0;
  for(int cont=0;cont<strlen(totest);cont++) {
    if(totest[cont]=='=') found=1;
    else if (found) h2+=totest[cont];
    else h1+=totest[cont];
  }
  *one=stringToChar(h1);
  *two=stringToChar(h2);
  return 1;

};

// realtodivide is divided in an operator (returned value) , a left expression
// (one)  and a right expression (two)
SI_operator divide(char* realtodivide, char** one,char** two) {

  string h1,h2,op;
  SI_operator toreturn;

#ifdef TWO_DEBUG_MODE
  cout << "divide: realtodivide e': " << realtodivide << "\n";
#endif

  unsigned int encountered=0;
  for(int cont=0;cont<strlen(realtodivide);cont++) {
    if(oprChar(realtodivide[cont])) {
      op+=realtodivide[cont];
      encountered=1;
    }
    else if(encountered) h2+=realtodivide[cont];
    else h1+=realtodivide[cont];
  }
  if ( (op[0]=='>') && (op.length()==1) ) toreturn=gt;
  if ( (op[0]=='<') && (op.length()==1) ) toreturn=lt;
  if ( (op[0]=='=') && (op.length()==1) ) toreturn=eq;
  if ( (op.length()==2) && (op[0]=='=') ) toreturn=leq;
  if ( (op.length()==2) && (op[0]=='>') ) toreturn=geq;
  if ( (op.length()==3) ) toreturn=neq;
  *one=stringToChar(h1);
  *two=stringToChar(h2);

#ifdef TWO_DEBUG_MODE
  cout<< "divide: h1 e' " << h1 << "\n";
  cout<< "divide: *one e' " << *one << "\n";
  cout<< "divide: h2 e' " << h2 << "\n";
  cout<< "divide: *two e' " << *two << "\n";
#endif

  return toreturn;

};

// tells whether the varname is matched in toparse starting from
// the specified position
unsigned int varPresent(char *varname,char *toparse,int position) {

  if( (position<0) || (strlen(toparse)-position<strlen(varname)) ) return 0;
  int cont;
  for(cont=0;cont<strlen(varname);cont++)
    { if(varname[cont]!=toparse[position+cont]) return 0; }
  unsigned int beginstr,endstr;
  beginstr=(position==0);
  endstr=(position+cont==strlen(toparse));
  if(beginstr && endstr) return 1;
  if(!beginstr && endstr)
    if(isLetter(toparse[position-1])) 
      return 0;
    else return 1;
  if(beginstr && !endstr)
    if(isLetter(toparse[cont+position]))
      return 0;
    else return 1;
  if(!beginstr && !endstr)
    if(isLetter(toparse[cont+position]) || isLetter(toparse[position-1]) )
      return 0;
    else return 1;

};

// tells whether the varname is present in the passed string
// no control on varname but the whole varname must be composed of
// capital letters
unsigned int varPresent(char *varname,char *toparse) {

  for(int cont=0;cont<strlen(toparse);cont++)
    { if(varPresent(varname,toparse,cont)) return 1; }
  return 0;

};

// returns the variables of the passed list that are present in the string
myvector<SIvariable> varsIn(char* toparse,myvector<SIvariable> tosearch) {

  tosearch.ResetCursor();
  myvector<SIvariable> toreturn;
  for(int cont=0;cont<tosearch.Length();cont++) {
    char* tmp=tosearch.CursorElement().getVar();
    if (varPresent(tmp,toparse))
      toreturn.Insert(tosearch.CursorElement());
    tosearch.IncCurs();
    delete [] tmp;
  }
  return toreturn;

};

// creates a  SIarithmetic from string and a variable list to be controlled
SIarithmetic getArithm(char* toparse,myvector<SIvariable> list) {

  SIarithmetic toreturn;
  myvector<SIvariable> toput;
  toput=varsIn(toparse,list);
  toput.ResetCursor();
  for(int cont=0;cont<toput.Length();cont++) {
    toreturn.addVariable(toput.CursorElement());
    toput.IncCurs();
  }
  toreturn.setExpr(toparse);
  return toreturn;

};

// Transforms the atom list ls into a formula
SIformula* build(atomlist ls) {

  SIformula *toreturn;
  toreturn=new SIformula();
  ls.resetCursor();
  if(ls.length()>0) 
    toreturn->setAtom(ls.cursorElement());
  ls.incCurs();
  for(int cont=1;cont<ls.length();cont++) {
    SIformula app;
    app.setAtom(ls.cursorElement());
    toreturn->SIand(app);
    ls.incCurs();
  }
  return toreturn;

};

// Transforms the dnf_vector ls into a formula
SIformula* build(dnf_vector ls) {

  SIformula *toreturn;
  SIformula *tmp_formula;
  toreturn=NULL;
  ls.resetCursor();
  if(ls.length()>0) 
    toreturn=build(ls.cursorElement());
  ls.incCurs();
  for(int  cont=1;cont<ls.length();cont++) {
    tmp_formula=build(ls.cursorElement());
    toreturn->SIor(*tmp_formula);
    delete tmp_formula;
    tmp_formula=NULL;
    ls.incCurs();
  }
  return toreturn;

};

// Performs the In Depth analysis over the real domain of toparse
// after the execution val is : 0 if the result is false, 1 if the result is
// is true, 2 if the result is neither true nor false
// If val is 2, the returned atomlist represents the simplified atomlist
atomlist realInDepth(atomlist toparse,int *val) {

  atomlist toreturn;

  toparse.resetCursor();
  for(int cont=0;cont<toparse.length();cont++) {
    toreturn.add(toparse.cursorElement());
    toparse.incCurs();
  }

  // premangling phase
  myvector<name_couple> ss=presymbolTable(toparse);
  premangle(&toparse,ss);

  // construct a symbol table for toparse and mangle it according to the table
  myvector<name_couple> rr=symbolTable(toparse);
  mangle(&toparse,rr);
  
  // get the new variables of toparse and put them in mm
  myvector<SIvariable> mm=vrs(toparse);
  
#ifdef TWO_DEBUG_MODE
  cout << "realInDepth chiama realQuery.\n";
  toparse.resetCursor();
  cout << "La lista passata e'  : \n";
  for(int cont11=0;cont11<toparse.length();cont11++) {
    
    char* tmp=toparse.cursorElement().toStringWithoutType();
    cout<< tmp << "\n";
    delete [] tmp;
    toparse.incCurs();

  }
#endif

  // Prepare a clpr query from toparse and send it to the prolog process
  char* query = realQuery(toparse);
  char* intVarList = getIntVarsToCommaSeparatedString(toparse);
  if (intVarList != NULL){
    query[strlen(query) - 1] = ',';
    string constrainedQuery = query;
    constrainedQuery += " bb_inf([";
    constrainedQuery += intVarList;
    constrainedQuery += "],0,_Inf).";
    delete [] query;
    query = new char[strlen(constrainedQuery.c_str()) + 1];
    strcpy(query, constrainedQuery.c_str());
    delete [] intVarList;
  }
  // send the query to the prolog process
  myvector<char *> answer=prologchannel->query(query);
  delete [] query;
  // If there is only one line in the answer the simplified list is identically
  // TRUE or identically FALSE
  if (answer.Length() == 1) { 
    string control;
    control=answer[0];
    if (control=="yes") { *val=1; return toreturn; }
    if (control=="no") { *val=0; return toreturn; }
  }
  // If the answer is longer than 1 line there are atoms in the simplified list
  else {
    while(toreturn.length()>0) { toreturn.cut(); }
    answer.ResetCursor();
    // Parse all the lines of prolog answer
    for(int cont=0; cont<answer.Length()-1; cont++) {
      
      // SIatom reconstruction procedure
      //
      // 1. Divide left and right expression from the prolog answer
      // 2. Get the comparison operator from the prolog answer
      // 3. Build an atom from the above information
      // 4. Add the atom in the list to be returned 
      char* left;
      char* right;
      // 1. - 2.
      SI_operator op=divide(answer.CursorElement(),&left,&right);

#ifdef TWO_DEBUG_MODE
      cout<< "realinDepth: left e': " << left << "\n";
      cout<< "realinDepth: right e': " << right << "\n";
#endif

      SIarithmetic one=getArithm(left,mm);
      SIarithmetic two=getArithm(right,mm);
      // 3.
      SIatom toadd;
      toadd.setAtom(one,two,op);
      // 4.
      toreturn.add(toadd);
      answer.IncCurs();
      delete [] left;
      delete [] right;
    }
  }
  // Brings back original variables names in the list to be returned
  demangle(&toreturn,rr);
  depremangle(&toreturn,ss);

  *val=2;
  // clears prolog answer
  answer.ResetCursor();
  for(int cont=0;cont<answer.Length();cont++) {
    char* toremove=answer.CursorElement();
    delete [] toremove;
    answer.IncCurs();
  }

  return toreturn;

};

// Performs the In Depth analysis over the integer domain of toparse  
// after the execution val is : 0 if the result is false, 1 if the result is
// is true, 2 if the result is neither true nor false
// The returned atom list contains a set of equality atoms expressing the
// value-bounded variables (and the value they are bound to)
atomlist integerInDepth(atomlist toparse,int *val) {

  atomlist toreturn;
  int cont;
  toparse.resetCursor();
  for(cont=0;cont<toparse.length();cont++) {
    toreturn.add(toparse.cursorElement());
    toparse.incCurs();
  }

  // prepare a premangle table for toparse
  myvector<name_couple> ss=presymbolTable(toparse);
  // premangle toparse according to rr
  premangle(&toparse,ss);

 
  // prepare a mangle table for toparse
  myvector<name_couple> rr=symbolTable(toparse);
  // mangle toparse according to rr
  mangle(&toparse,rr);


  // stores the new variables of toparse in mm
  myvector<SIvariable> mm=vrs(toparse);

#ifdef TWO_DEBUG_MODE

  cout<< "integerInDepth: sto passando a integerQuery una lista di dimensione: ";

  if(toparse.length()>0) 
    cout << ">0.\n";
  else cout << "=0!!!\n"; 
  cout<< "La lista che sto passando a integerequery:\n";
  toparse.resetCursor();
  for(int cont3=0;cont3<toparse.length();cont3++) {
    char* tmp=toparse.cursorElement().toStringWithoutType();
    cout<< tmp << "\n";
    delete [] tmp;
    toparse.incCurs();
  }
  
#endif

  // prepare a Sicstus clpfd query from toparse
  char* query=integerQuery(toparse);
  // send the query to the prolog process
  myvector<char *> answer=prologchannel->query(query);
  delete [] query;
  // If there is only one line in the answer the simplified list is identically
  // TRUE or identically FALSE

  if (answer.Length()==1) {
    string truthanswer;
    truthanswer=answer[0];
    if (truthanswer=="yes") { *val=1; return toreturn; }
    if (truthanswer=="no") { *val=0; return toreturn; }
  }
  // If the answer is longer than 1 line there are atoms in the simplified list
  else {
    answer.ResetCursor();
    // Parse all the lines of prolog answer
    for(cont=0;cont<answer.Length()-1;cont++) {
      // SIatom reconstruction procedure
      //
      // 1. Divide left and right expression from the prolog answer
      // (the comparison operator is always " = ", since the integer 
      // simplification
      // step simply detects when one or more variables are bound to a specific
      // value)
      // 2. Build an atom from the above information
      // 3. Add the atom in the list to be returned 
      char* left=NULL;
      char* right=NULL;
      // 1.
      if(divideEq(answer.CursorElement(),&left,&right)){
#ifdef TWO_DEBUG_MODE
	cout<< "integerInDepth: left e': " << left << "\n";
	cout<< "integerInDepth: right e': " << right << "\n";
#endif
	// 2.
	SIarithmetic one=getArithm(left,mm);
	SIarithmetic two=getArithm(right,mm);
	SIatom toinsert;
	toinsert.setAtom(one,two,eq);
	// 3.
	toparse.add(toinsert);
      }
      answer.IncCurs();
      delete [] left;
      delete [] right;
    }
    

    // bring back toparse original variables
    demangle(&toparse,rr);
    depremangle(&toparse,ss);

 
    while(toreturn.length()>0) { toreturn.cut(); }
    // Add the equality atoms deriving from integer simplification to the atom
    // list to be returned
    toparse.resetCursor();
    for(cont=0;cont<toparse.length();cont++) {
      toreturn.add(toparse.cursorElement());
      toparse.incCurs();
    }
    int fake;
    
#ifdef TWO_DEBUG_MODE
    cout << "integerInDepth chiama realInDept.\n";
    cout << "Ecco la lista che sto per passare a realInDepth da integerInDepth:\n";
    toreturn.resetCursor();
    for(int cont7=0;cont7<toreturn.length();cont7++) {
      char* tmp=toreturn.cursorElement().toStringWithoutType();
      cout<< tmp << "\n";
      delete [] tmp; 
      toreturn.incCurs();
    }
#endif
        
    // Perform an In Depth analysis over the real domain of the atom list to be
    // returned; toreturn now contains the equality atoms representing the
    // value-bounded variables deriving from the integer simplification step
    // (they are added in logical AND conjunction) 
    toreturn=realInDepth(toreturn,&fake);
    *val=2;
    answer.ResetCursor();
    for(int cont=0;cont<answer.Length();cont++) {
      char* toremove=answer.CursorElement();
      delete [] toremove;
      answer.IncCurs();
    }
    return toreturn;
  }

};


// Performs the In Depth analysis over the integer domain of toparse  
// after the execution val is : 0 if the result is false, 1 if the result is
// is true, 2 if the result is neither true nor false
// The returned atom list contains a set of equality atoms expressing the
// value-bounded variables (and the value they are bound to)
// this variant uses multimangle tables for mixed atom lists
atomlist integerInDepthForMixed(atomlist toparse,int *val) {

  atomlist toreturn;
  atomlist intlist;
  unsigned int allIntAtoms=1;
  int cont;
  toparse.resetCursor();
  for(cont=0;cont<toparse.length();cont++) {
    //seleziono gli atomi con variabili solo intere
    unsigned int atomtype = typeRecognizeAtom(toparse.cursorElement());
#ifdef TWO_DEBUG_MODE
    cout<< "integerInDepthForMixed: analizzando l'atomo: \n";
    char* tmp=toparse.cursorElement().toStringWithoutType();
    cout<< tmp << ",\n";
    delete [] tmp;
    cout<<"typeRecognizeAtom returns type (1=int, 2=real, 3=mixed): "
	     << atomtype<<"\n";
#endif
    
    if(atomtype==1) intlist.add(toparse.cursorElement());
    else {
      toreturn.add(toparse.cursorElement());
      allIntAtoms=0;
    }
    toparse.incCurs();
  }

  if(intlist.length()==0){
    *val=2;
    return toreturn;
  }
#ifdef TWO_DEBUG_MODE
  cout<< "integerInDepthForMixed: sto passando a integerQuery una lista di dimensione: " << intlist.length() << "\n";
  cout<< "La lista e':\n";
  intlist.resetCursor();
  for(int cont3=0;cont3<intlist.length();cont3++) {
    char* tmp=intlist.cursorElement().toStringWithoutType();
    cout<< tmp << "\n";
    delete [] tmp;
    intlist.incCurs();
  }
#endif

  // prepare a premangle table for intlist
  myvector<name_couple> ss=presymbolTable(intlist);
  // premangle intlist according to rr
  premangle(&intlist,ss);

#ifdef TWO_DEBUG_MODE

  cout<< "La lista dopo il premangling:\n";
  intlist.resetCursor();
  for(int cont3=0;cont3<intlist.length();cont3++) {
    char* tmp=intlist.cursorElement().toStringWithoutType();
    cout<< tmp << "\n";
    delete [] tmp;
    intlist.incCurs();
  }
#endif

  // prepare a mangle table for intlist
  // ints indicates how many multimangle entries are in the symbol table
  unsigned int ints;
  myvector<name_couple> rr=symbolTableForMixed(intlist,&ints);
  // mangle intlist according to rr
  mangleForMixed(&intlist,rr,ints);

#ifdef TWO_DEBUG_MODE

  cout<< "La lista dopo il mangling for mixed:\n";
  intlist.resetCursor();
  for(int cont3=0;cont3<intlist.length();cont3++) {
    char* tmp=intlist.cursorElement().toStringWithoutType();
    cout<< tmp << "\n";
    delete [] tmp;
    intlist.incCurs();
  }
#endif

  // stores the new variables of intlist in mm
  myvector<SIvariable> mm=vrs(intlist);

#ifdef TWO_DEBUG_MODE

  cout<< "integerInDepthforMixed: sto passando a integerQuery una lista di dimensione: " << intlist.length() << "\n";

  cout<< "La lista che sto passando a integerequery:\n";
  intlist.resetCursor();
  for(int cont3=0;cont3<intlist.length();cont3++) {
    char* tmp=intlist.cursorElement().toStringWithoutType();
    cout<< tmp << "\n";
    delete [] tmp;
    intlist.incCurs();
  } 
#endif

  // prepare a Sicstus clpfd query from intlist
  char* query=integerQuery(intlist);
  // send the query to the prolog process
  myvector<char *> answer=prologchannel->query(query);
  delete [] query;

  // If there is only one line in the answer, the simplified list is
  // identically TRUE or identically FALSE. A FALSE value makes FALSE
  // the all conjunction. In case of TRUE value the answer depends on
  // the possible real atoms. If there was no real atom otherwise the
  // value of the whole formula is TRUE , otherwise only the list of
  // real atoms is returned
  if (answer.Length()==1){
    string truthanswer;
    truthanswer=answer[0];
    if (truthanswer=="no") *val=0; 
    else if (truthanswer=="yes") { 
      if(allIntAtoms) *val=1; 
      else *val=2;
    }
    else {
      printf("Assert:integerInDepthForMixed:unexpected answer from sicstus\n");
      exit(-1);
    }
    return toreturn;
  }
  // else: If the answer is longer than 1 line there are atoms in the simplified list
  answer.ResetCursor();
  // Parse all the lines of prolog answer
  for(cont=0;cont<answer.Length();cont++) {
 // SIatom reconstruction procedure
 //
 // 1. Divide left and right expression from the prolog answer
 // (the comparison operator is always " = ", since the integer simplification
 // step simply detects when one or more variables are bound to a specific
 // value
 // 2. Build an atom from the above information
 // 3. Add the atom in the list to be returned 
    char* left=NULL;
    char* right=NULL;
    // 1.
    if(divideEq(answer.CursorElement(),&left,&right)){
#ifdef TWO_DEBUG_MODE
      cout<< "integerInDepthForMixed: left e': " << left << "\n";
      cout<< "integerInDepthForMixed: right e': " << right << "\n";
#endif
	// 2.
	SIarithmetic one=getArithm(left,mm);
	SIarithmetic two=getArithm(right,mm);
	SIatom toinsert;
	toinsert.setAtom(one,two,eq);
      // 3.
      intlist.add(toinsert);
    }
    delete [] left;
    delete [] right;
    answer.IncCurs();
  }

  // bring back toparse original variables
  demangle(&intlist,rr);
  depremangle(&intlist,ss);

  //while(toreturn.length()>0) { toreturn.cut(); }
  
  //Add the equality atoms deriving from integer simplification to the atom
  // list to be returned
  intlist.resetCursor();
  for(cont=0;cont<intlist.length();cont++) {
    toreturn.add(intlist.cursorElement());
    intlist.incCurs();
  }
  int fake;

#ifdef TWO_DEBUG_MODE
  cout << "integerInDepth chiama realInDept.\n";
  cout << "Ecco la lista che sto per passare a realInDepth da integerInDepth:\n";
  toreturn.resetCursor();
  for(int cont7=0;cont7<toreturn.length();cont7++) {
      char* tmp=toreturn.cursorElement().toStringWithoutType();
      cout<< tmp << "\n";
      delete [] tmp; 
      toreturn.incCurs();
   }
#endif
        
    // Perform an In Depth analysis over the real domain of the atom list to be
    // returned; toreturn now contains the equality atoms representing the
    // value-bounded variables deriving from the integer simplification step
    // (they are added in logical AND conjunction) 
    toreturn=realInDepth(toreturn,&fake);
    *val=2;
    answer.ResetCursor();
    for(int cont=0;cont<answer.Length();cont++) {
      char* toremove=answer.CursorElement();
      delete [] toremove;
      answer.IncCurs();
    }
						  
  return toreturn;

};


// Performs In Depth analysis of the atomlist toparse
// This method is meant to analyze atom lists that contain only integer variables
// after the execution val is : 0 if the result is false, 1 if the result is
// is true, 2 if the result is neither true nor false
atomlist inDepthForMixed(atomlist toparse,int *val) {

  atomlist toreturn;
  int value;
  
#ifdef TWO_DEBUG_MODE
  cout << "inDepthForMixed chiama realInDepth.\n";
  cout << "Ecco la lista che sto per passare a realInDepth:\n";
  toparse.resetCursor();
  for(int cont7=0;cont7<toparse.length();cont7++) {
    char* tmp=toparse.cursorElement().toStringWithoutType();
    cout<<  tmp << "\n"; 
    delete [] tmp;
    toparse.incCurs();
  }
#endif

  // Simplification over the real domain
  toreturn=realInDepth(toparse,&value);

#ifdef TWO_DEBUG_MODE
  cout << "inDepthForMixed riceve risultato da realInDepth.\n";
  cout << "Ecco la lista che ricevuta da realInDepth:\n";
  toreturn.resetCursor();
  for(int cont7=0;cont7<toreturn.length();cont7++) {
    char* tmp=toreturn.cursorElement().toStringWithoutType();
    cout<<  tmp << "\n"; 
    delete [] tmp;
    toreturn.incCurs();
  }
  if (toreturn.length()>0) cout << "inDepthForMixed: la lunghezza della atomlist da mandare a Sicstus e' >0\n";
  else cout << "La lunghezza della lista da mandare a Sicstus e' 0 !!\n";
#endif
  // If the result of simplification over real domain is neither TRUE nor FALSE,
  // perform the simplification over the integer domain step
  if (value==2) {
    toreturn=integerInDepthForMixed(toreturn,&value);
  }
  *val=value;

  return toreturn;

};


// Performs In Depth analysis of the atomlist toparse
// This method is meant to analyze atom lists that contain only integer variables
// after the execution val is : 0 if the result is false, 1 if the result is
// is true, 2 if the result is neither true nor false
atomlist inDepthForInteger(atomlist toparse,int *val) {

  atomlist toreturn;
  int value;
  
#ifdef TWO_DEBUG_MODE
  cout << "inDepthForInteger chiama realInDepth.\n";
  cout << "Ecco la lista che sto per passare a realInDepth:\n";
  toparse.resetCursor();
  for(int cont7=0;cont7<toparse.length();cont7++) {
    char* tmp=toparse.cursorElement().toStringWithoutType();
    cout<<  tmp << "\n"; 
    delete [] tmp;
    toparse.incCurs();
  }
#endif

  // Simplification over the real domain
  toreturn=realInDepth(toparse,&value);

#ifdef TWO_DEBUG_MODE
  if (toreturn.length()>0) cout << "inDepthForInteger: la lunghezza della atomlist da mandare a Sicstus e' >0\n";
  else cout << "La lunghezza della lista da mandare a Sicstus e' 0 !!\n";
#endif
  // If the result of simplification over real domain is neither TRUE nor FALSE,
  // perform the simplification over the integer domain step
  if (value==2) {
    toreturn=integerInDepth(toreturn,&value);
  }
  *val=value;

  return toreturn;

};


// Performs In Depth analysis of the atomlist toparse
// This method is meant to analyze atom lists that contain both  integer and real variables
// after the execution val is : 0 if the result is false, 1 if the result is
// is true, 2 if the result is neither true nor false
atomlist inDepthForReal(atomlist toparse,int *val) {

  atomlist toreturn;
  int value;
  
#ifdef TWO_DEBUG_MODE
  cout << "inDepthForReal chiama realInDepth.\n";
  cout << "Ecco la lista che sto per passare a realInDepth:\n";
  toparse.resetCursor();
  for(int cont7=0;cont7<toparse.length();cont7++) {
    char* tmp=toparse.cursorElement().toStringWithoutType();
    cout<<  tmp << "\n"; 
    delete [] tmp;
    toparse.incCurs();
  }
#endif

  // Simplification over the real domain
  toreturn=realInDepth(toparse,&value);

#ifdef TWO_DEBUG_MODE
  if (toreturn.length()>0) cout << "inDepthforReal: la lunghezza della atomlist da mandare a Sicstus e' >0\n";
  else cout << "La lunghezza della lista da mandare a Sicstus e' 0 !!\n";
#endif

  *val=value;

  return toreturn;

};


// Performs In Depth analysis of the atomlist toparse
// after the execution val is : 0 if the result is false, 1 if the result is
// is true, 2 if the result is neither true nor false
atomlist inDepth(atomlist toparse,int *val) {
  return realInDepth(toparse,val);//inDepthForReal(toparse,val);

    /* unsigned int type;
    type=typeRecognizeAtomlist(toparse);
#ifdef TWO_DEBUG_MODE
    cout << "inDepth: recognized expression of type(1=int, 2=real, 3=mixed):"<<type<<"\n";
#endif
    if( (type==0) || (type==2) ) return inDepthForReal(toparse,val);
    if(type==1) return inDepthForInteger(toparse,val);
    else return inDepthForMixed(toparse,val);
    */
};


// Tells whether p is contained in t
unsigned int listContained(atomlist p,dnf_vector t) {

  t.resetCursor();
  for(int cont=0;cont<t.length();cont++) {
    if(p==t.cursorElement()) return 1;
    t.incCurs();
  }
  return 0;
  
};


// Tells whether the two arithmetic expressions are equivalent by querying sicstus
unsigned int equalExpressions(SIarithmetic a,SIarithmetic b) {

  SIatom *testat;
  atomlist *testlist;
  int truthcontrol=0;

  testat=new SIatom(a,b,eq);
  testlist=new atomlist();
  testlist->add(*testat);
  realInDepth(*testlist,&truthcontrol);
  delete testat;
  delete testlist;
  if (truthcontrol==1) return 1;
  else return 0;

};


// Performs inWidthAnalysis of tosimplify
// t is 1 if the dnf vector simplifies to true
dnf_vector inWidthAnalysis(dnf_vector tosimplify,unsigned int *t) {

  dnf_vector toreturn;
  dnf_vector copy;
  dnf_vector app;
  *t=0;
  
  tosimplify.resetCursor();
  for (int cont=0;cont<tosimplify.length();cont++) {
    if (!listContained(tosimplify.cursorElement(),app)) {  
      app.insert(tosimplify.cursorElement());
      copy.insert(tosimplify.cursorElement());
    }
    tosimplify.incCurs();
  }

  app.resetCursor();
  for(int cont=0;cont<app.length();cont++) {
    copy.resetCursor();
    unsigned int toput=1;
    // Inclusion check: the current atom list will be added to toreturn (toput=1)
    // only if there are no lists that contain it
    for(int cont2=0;cont2<copy.length();cont2++) { 
      if( app.cursorElement() > copy.cursorElement() ) {  
	toput=0;
	break;
      }
      copy.incCurs();
    }
    if(toput) toreturn.insert(app.cursorElement());
    app.incCurs();
  }

  return toreturn;
  
};

// Performs In Depth analysis of tosimplify
dnf_vector inDepthAnalysis(dnf_vector tosimplify) {

  dnf_vector toreturn;
  tosimplify.resetCursor();
  unsigned int istrue=0;
  unsigned int isfalse=0;
  tosimplify.resetCursor();
  // If tosimplify contains only one list
  // if the list is true the whole dnf_vector is true
  // if the list is false the whole dnf_vector is false
  if (tosimplify.length()==1) {
    if (tosimplify.cursorElement().cursorElement().isTrue()) {
      SIatom toadd;
      toadd.setLogicValue(1);
      atomlist truelist;
      truelist.add(toadd);
      toreturn.add(truelist);
      return toreturn;
    }
    if (tosimplify.cursorElement().cursorElement().isFalse()) {
      SIatom toadd;
      toadd.setLogicValue(0);
      atomlist falselist;
      falselist.add(toadd);
      toreturn.add(falselist);
      return toreturn;
    }
  }
  // If the whole dnf_vector does not simplify to TRUE or to FALSE, all its
  // atom lists must be checked one by one
  for(int cont=0;cont<tosimplify.length();cont++) {
    int val;
    atomlist element;
    element=inDepth(tosimplify.cursorElement(),&val);
    // if val=0 (FALSE atomlist) the atom list is no more added to toreturn
    // if val=1 the whole dnf_vector becomes TRUE (see below)
    // if val=2 the returned atom list is added to toreturn
    if(val==2) toreturn.add(element);
    if(val==1) istrue=1;
    tosimplify.incCurs();
  }
  // the whole dnf_vector becomes TRUE
  if (istrue) {
    while(toreturn.length()>0) { toreturn.cut(); }
    SIatom toadd;
    toadd.setLogicValue(1);
    atomlist truelist;
    truelist.add(toadd);
    toreturn.add(truelist);
  }
  return toreturn;

};


/**************************** END OF SUPPORT METHODS ******************************/





/***************************** SIMPLIFIER EXCEPTION *******************************/

void SIsimplerr::setErr(unsigned int toset) {

  err=toset;

};

unsigned int SIsimplerr::getErr() {

  return err;

};

/************************* END OF SIMPLIFIER EXCEPTION ***************************/






/********************** SIsimplifier INTERFACE METHODS **************************/



// throws an exception if the initialization fails
SIsimplifier::SIsimplifier(unsigned int slot) {
  
  unsigned int ok;
  initialized=0;
  ok=iface.initialize(slot);
  prologchannel=&iface;
  if(!ok) {
    SIsimplerr exc;
    exc.setErr(0);
    throw(exc);
  }
  else initialized=1;

};

SIsimplifier::~SIsimplifier() {
  iface.closeProlog();
  initialized=0;
};

/******************************************************
 *                                                    *
 *         SIMPLIFICATION SUPPORT FUNCTION            *
 *                                                    *
 ******************************************************/                 

void SIsimplifier::simplifyF(SIformula *tosimplify) {

  // if (tosimplify->istrue || tosimplify->isfalse)
  //  return; //do nothing
  
  // Excpetion raised if the method is called when the instance is not initialized

  if(!initialized) {
    SIsimplerr tr;
    tr.setErr(1);
    throw(tr);
  }

  dnf_vector totest;
  
  // If tosimplify is TRUE or FALSE simply return a TRUE or a FALSE formula
  if(tosimplify->istrue) {
    SIatom toadd;
    toadd.setLogicValue(1);
    atomlist truelist;
    truelist.add(toadd);
    totest.add(truelist);
  }

  if(tosimplify->isfalse) {
    SIatom toadd;
    toadd.setLogicValue(0);
    atomlist falselist;
    falselist.add(toadd);
    totest.add(falselist);
  }

  // If tosimplify is neither TRUE nor FALSE perform In Depth an In Width analysis


  if ( (!tosimplify->isfalse) && (!tosimplify->istrue) ) 

    // Get the DNF of tosimplify
    totest=tosimplify->dnForm();

  // Prepare the DNF to be returned
  dnf_vector result;
  // In Depth analysis
  result=inDepthAnalysis(totest);
  SIformula *simplified;
  // If In Depth analysis returns only FALSE atomlists the whole formula is FALSE
  if(result.length()==0) {
    simplified=new SIformula();
    simplified->setLogicVal(0);
  }
  // If In Depth analysis returns only one atom list the formula is
  // TRUE if the atom list is TRUE
  // built directly from the atom list otherwise
  if(result.length()==1) {

    if(result.cursorElement().cursorElement().isTrue()) {
      simplified=new SIformula();
      simplified->setLogicVal(1);
    }
    else {
      unsigned int truthcontrol;
      result=inWidthAnalysis(result,&truthcontrol);
      if(truthcontrol) {
	simplified=new SIformula();
	simplified->setLogicVal(1);
      }
      else simplified=build(result);
    }

  }

  // If In Depth analysis returns more than 1 list In Width analysis must be
  // performed
  if(result.length()>1) {

    unsigned int truthcontrol;
    result=inWidthAnalysis(result,&truthcontrol);
    // If truthcontrol=1 In Width analysis has detected a TRUE formula 
    if(truthcontrol) {
      simplified=new SIformula();
      simplified->setLogicVal(1);
    }
    // otherwise the simplified formula can be built directly from result 
    else simplified=build(result);
  }

  // Destroy tosimplify to rebuild it by copying the simplified formula
  delete(tosimplify->top);
  tosimplify->istrue=simplified->istrue;
  tosimplify->isfalse=simplified->isfalse;
  if( (!tosimplify->istrue) && (!tosimplify->isfalse) ) {
    tosimplify->top=new formulanode(*simplified->top);
    tosimplify->top->dnform=simplified->top->dnform;
    tosimplify->top->setNext(NULL);
    tosimplify->firstatom=tosimplify->firstAtom(tosimplify->top);
    tosimplify->lastatom=tosimplify->lastAtom(tosimplify->top);
  }
  else
    tosimplify->top=NULL;
  
  // SIformula simplified is no more needed (already copied in tosimplify)
  delete simplified;
};

void SIsimplifier::simplify(SIformula *tosimplify) {

  SIformula *negation;
  negation=new SIformula(*tosimplify);
  negation->SInot();
  //GIOVANNI 6-8-2008 NB: 2 steps of simplification are necessary
  //here, due to the current implementation that calls sicstus once
  //for each disjunct in the DNF (disjunctive normal form) of the
  //formula. If each disjunctive is satisfiable, we know that the
  //formula is satisfiable but yet do not know whether the negation of
  //the formula is satisfiable too. The viceversa holds if we would
  //just test the negated formula.  Note that since our formulas are
  //conjunctions of PC terms, the DNF of the negated formula is the
  //disjunction of all PC terms, so that sicstus get invoked for each
  //term separately.
  simplifyF(tosimplify);
  simplifyF(negation);
  if(negation->isTrue()) tosimplify->setLogicVal(0);
  if(negation->isFalse()) tosimplify->setLogicVal(1);
  delete negation;
 
};


// Performs the propagation of the equality AND conjunction simplifier
// over the formula tosimplify
// returns 3 if simplifier forces  tosimplify to false
// returns 0 if simplifier doesn't affect tosimplify (no subst.)
// returns 1 if simplifier affects tosimplify but doesn't force it to false
// returns 2 if simplifier is not a postcondition (nothing to  be done)
int SIsimplifier::propagate(SIformula *tosimplify,SIformula &simplifier) {

  // Excpetion raised if the method is called when the instance is not initialized
  if(!initialized) {
    SIsimplerr tr;
    tr.setErr(1);
    throw(tr);
  }

  if (tosimplify->isTrue() || simplifier.isTrue())
    return 0;
  if (!simplifier.eqConj()) return 2;
  

  formulanode *parser;

  
  // Substitutes all variables with apices with variables without primes
  // in tosimplify by creating a substitution table (subsvrs)
  myvector<SIvariable> curvrs=tosimplify->getVariables();
  myvector<name_couple> subsvrs;
  curvrs.ResetCursor();
  for(int cont=0;cont<curvrs.Length();cont++) {
    name_couple nc;
    char *tmp;
    tmp=curvrs.CursorElement().getVar();
    nc.setOne(tmp);
    string control=tmp;

#ifdef TWO_DEBUG_MODE
    if (control[control.length()-1]!='\'') 
      cout << "Variabile senza apice in una condizione d'esecuzione!\n";
#endif

    delete [] tmp;
    string jj;
    jj=control.substr(0,control.length()-1);
    tmp=stringToChar(jj);
    nc.setTwo(tmp);
    curvrs.IncCurs();
    subsvrs.Insert(nc);
    delete [] tmp;
  }
  
  // By using the substitution table subsvrs constructed above, eliminates
  // all the primes from the variables names of tosimplify
  subsvrs.ResetCursor();
  for(int cont=0;cont<subsvrs.Length();cont++) {
    char *tmp,*tmp2;
    tmp=subsvrs.CursorElement().getOne();
    tmp2=subsvrs.CursorElement().getTwo();
    tosimplify->replace(tmp,tmp2);
    delete [] tmp;
    delete [] tmp2;
    subsvrs.IncCurs();
  } 
  
  // Perform substitution dictated by simplifier formula
  //
  // 1. Create the propagation table (proptable) : a table telling which
  //    expression to substitute for each variable
  // 2. Use proptable to perform the substitutions dictated by simplifier
  
  // 1.
  myvector<varexpr_couple> proptable;

  parser=simplifier.firstatom;

  while (parser!=NULL) {
    varexpr_couple toput;
    char* tmp=parser->atm.leftExpr().toStringWithoutType();
    toput.setOne(tmp);
    toput.setTwo(parser->atm.rightExpr());
    proptable.Insert(toput);
    parser=parser->nextatom;
    delete [] tmp;
  }

  // 2.
  proptable.ResetCursor();
  unsigned int affected=0;
  for(int cont=0;cont<proptable.Length();cont++) {
    
    char* tmp2=proptable.CursorElement().getOne();
    
#ifdef TWO_DEBUG_MODE
    char* tmp=proptable.CursorElement().getTwo().toStringWithoutType();
    cout << "Elemento della proptable : " << tmp2 << " " << tmp << "\n";
    
    delete [] tmp;
    myvector<SIvariable> prnt= proptable.CursorElement().getTwo().getVariables();
    
    cout << "Variabili secondo elemento della coppia: \n";
    prnt.ResetCursor();
    for(int cont2=0;cont2<prnt.Length();cont2++) {
      cout << prnt.CursorElement().getVar() << "\n";
      prnt.IncCurs();
    }
#endif

    if (tosimplify->replace(tmp2,proptable.CursorElement().getTwo())) 
      affected=1;
    delete [] tmp2;
    proptable.IncCurs();
  }
  
  // Re-brings simbols with primes back using subsvrs

  subsvrs.ResetCursor();
  for(int cont=0;cont<subsvrs.Length();cont++) {
    char *tmp,*tmp2;
    tmp2=subsvrs.CursorElement().getTwo();
    tmp=subsvrs.CursorElement().getOne();
    
    tosimplify->replace(tmp2,tmp);
    delete [] tmp;
    tmp=NULL;
    delete [] tmp2;
    tmp2=NULL;
    subsvrs.IncCurs();
  }   
  

  // If simplifier does not affect tosimplify return 0
  if (!affected) return 0;
  // otherwhise simplify tosimplify
  simplify(tosimplify);
  if(tosimplify->isFalse()) return -1;
  else return 1;

};

/****************** END OF SIsimplifier INTERFACE METHODS *************************/
